# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

from unittest.mock import patch

import balance.testutil

import glmnet_python  # noqa  # Required so that cvglmnet import works
import numpy as np
import pandas as pd

from balance.sample_class import Sample
from balance.weighting_methods import ipw as balance_ipw

from cvglmnet import cvglmnet  # pyre-ignore[21]: this module exists
from cvglmnetCoef import cvglmnetCoef  # pyre-ignore[21]: this module exists


class Testipw(
    balance.testutil.BalanceTestCase,
):
    def test_ipw_weights_order(self):
        sample = pd.DataFrame({"a": (1, 2, 3, 4, 5, 6, 7, 9, 1)})
        target = pd.DataFrame({"a": (1, 2, 3, 4, 5, 6, 7, 9, 9)})

        result = balance_ipw.ipw(
            sample_df=sample,
            sample_weights=pd.Series((1,) * 9),
            target_df=target,
            target_weights=pd.Series((1,) * 9),
            transformations=None,
            max_de=1.5,
        )

        w = result["weight"].values

        self.assertEqual(w[0], w[8])
        self.assertTrue(w[0] < w[1])
        self.assertTrue(w[0] < w[7])

    def test_ipw_sample_indicator(self):
        # TODO: verify this behavior of balance is sensible

        # error message
        # TODO: is this test working?
        s = pd.DataFrame({"a": np.random.uniform(0, 1, 1000), "id": range(0, 1000)})
        t = pd.DataFrame({"a": np.random.uniform(0.5, 1.5, 1000), "id": range(0, 1000)})
        s_w = pd.Series(np.array((1,) * 1000))
        t_w = pd.Series(np.array((1,) * 1000))
        self.assertRaises(
            Exception, "same number of rows", balance_ipw.ipw, s, s_w, t, t_w
        )

        t = pd.DataFrame({"a": np.random.uniform(0.5, 1.5, 999), "id": range(0, 999)})
        # Doesn't raise an error
        balance_ipw.ipw(
            sample_df=s,
            sample_weights=s_w,
            target_df=t,
            target_weights=pd.Series(np.array((1,) * 999)),
        )

    def test_ipw_bad_adjustment_warnings(self):
        #  Test the check for identical weights
        #  Test the check for no large coefficients
        #  Test the check for model accuracy
        n = 100
        sample = Sample.from_frame(
            df=pd.DataFrame(
                {"a": np.random.normal(0, 1, n).reshape((n,)), "id": range(0, n)}
            ),
            id_column="id",
        )
        sample = sample.set_target(sample)
        self.assertWarnsRegexp(
            "All weights are identical. The estimates will not be adjusted",
            sample.adjust,
            method="ipw",
            balance_classes=False,
        )
        self.assertWarnsRegexp(
            (
                "All propensity model coefficients are zero, your covariates do "
                "not predict inclusion in the sample. The estimates will not be "
                "adjusted"
            ),
            sample.adjust,
            method="ipw",
            balance_classes=False,
        )
        self.assertWarnsRegexp(
            (
                "The propensity model has low fraction null deviance explained "
                ".*. Results may not be accurate"
            ),
            sample.adjust,
            method="ipw",
            balance_classes=False,
        )

    def test_ipw_na_drop(self):
        s = Sample.from_frame(
            pd.DataFrame(
                {
                    "a": np.concatenate(
                        (np.array([np.nan, np.nan]), np.arange(3, 100))
                    ),
                    "id": np.arange(1, 100),
                }
            ),
            id_column="id",
        )

        t = Sample.from_frame(
            pd.DataFrame(
                {
                    "a": np.concatenate((np.array([np.nan]), np.arange(2, 100))),
                    "id": np.arange(1, 100),
                }
            ),
            id_column="id",
        )

        self.assertWarnsRegexp(
            "Dropped 2/99 rows of sample",
            s.adjust,
            t,
            na_action="drop",
            transformations=None,
        )

    def test_cv_glmnet_performance(self):
        x = np.random.rand(100, 10)
        y = np.random.rand(100, 1)
        y = (y > 0.5) * 1.0
        with balance_ipw._patch_scipy_random():
            fit = cvglmnet(x=x, y=y, family="binomial", ptype="class")

        #  Test default args with feature names
        c = cvglmnetCoef(fit)
        e_c_names = pd.Series(
            index="intercept a b c d e f g h i j".split(), data=c.reshape((c.shape[0],))
        )
        e_c = pd.Series(data=c.reshape((c.shape[0],)))

        r = balance_ipw.cv_glmnet_performance(
            fit, feature_names="a b c d e f g h i j".split()
        )

        ix = fit["lambdau"] == fit["lambda_1se"]
        e_dev = fit["glmnet_fit"]["dev"][ix]
        e_cve = fit["cvm"][ix]

        self.assertEqual(r["coefs"], e_c_names)
        self.assertEqual(r["prop_dev_explained"], e_dev)
        self.assertEqual(r["mean_cv_error"], e_cve)

        self.assertEqual(len(fit["glmnet_fit"]["dev"].shape), 1)
        self.assertEqual(len(fit["cvm"].shape), 1)

        #  Test default args with lambda_min
        c = cvglmnetCoef(fit, s="lambda_min")
        e_c_names = pd.Series(
            index="intercept a b c d e f g h i j".split(), data=c.reshape((c.shape[0],))
        )
        e_c = pd.Series(data=c.reshape((c.shape[0],)))
        r = balance_ipw.cv_glmnet_performance(fit, s="lambda_min")

        ix = fit["lambdau"] == fit["lambda_min"]
        e_dev = fit["glmnet_fit"]["dev"][ix]
        e_cve = fit["cvm"][ix]

        self.assertEqual(r["coefs"], e_c)
        self.assertEqual(r["prop_dev_explained"], e_dev)
        self.assertEqual(r["mean_cv_error"], e_cve)

        #  Test with specific lambda
        s = np.array([fit["lambdau"][5]])
        c = cvglmnetCoef(fit, s=s)
        e_c_names = pd.Series(
            index="intercept a b c d e f g h i j".split(), data=c.reshape((c.shape[0],))
        )
        e_c = pd.Series(data=c.reshape((c.shape[0],)))
        r = balance_ipw.cv_glmnet_performance(fit, s=s)

        ix = fit["lambdau"] == s
        e_dev = fit["glmnet_fit"]["dev"][ix]
        e_cve = fit["cvm"][ix]

        self.assertEqual(r["coefs"], e_c)
        self.assertEqual(r["prop_dev_explained"], e_dev)
        self.assertEqual(r["mean_cv_error"], e_cve)

        # Test exception
        balance_ipw.cv_glmnet_performance(fit, s=s)
        self.assertRaisesRegex(
            Exception,
            "No lambda found for",
            balance_ipw.cv_glmnet_performance,
            fit,
            s=7,
        )

    def test_weights_from_link(self):
        link = np.array((1, 2, 3)).reshape(3, 1)
        target_weights = (1, 2)
        r = balance_ipw.weights_from_link(link, False, (1, 1, 1), target_weights)
        e = np.array((1 / np.exp(1), 1 / np.exp(2), 1 / np.exp(3)))
        e = e * np.sum(target_weights) / np.sum(e)
        self.assertEqual(r, e)
        self.assertEqual(r.shape, (3,))

        # balance_classes does nothing if classes have same sum weights
        r = balance_ipw.weights_from_link(link, True, (1, 1, 1), (1, 2))
        self.assertEqual(r, e)

        # balance_classes uses link+log(odda)
        target_weights = (1, 2, 3)
        r = balance_ipw.weights_from_link(
            link, True, (1, 1, 1), target_weights, keep_sum_of_weights=False
        )
        e = np.array(
            (
                1 / np.exp(1 + np.log(1 / 2)),
                1 / np.exp(2 + np.log(1 / 2)),
                1 / np.exp(3 + np.log(1 / 2)),
            )
        )
        e = e * np.sum(target_weights) / np.sum(e)
        self.assertEqual(r, e)

        # sample_weights doubles
        target_weights = (1, 2)
        r = balance_ipw.weights_from_link(link, False, (2, 2, 2), target_weights)
        e = np.array((2 / np.exp(1), 2 / np.exp(2), 2 / np.exp(3)))
        e = e * np.sum(target_weights) / np.sum(e)
        self.assertEqual(r, e)

        # trimming works
        r = balance_ipw.weights_from_link(
            np.random.uniform(0, 1, 10000),
            False,
            (1,) * 10000,
            (1),
            weight_trimming_percentile=(0, 0.11),
            keep_sum_of_weights=False,  # this parameter is passed to trim_weights
        )
        self.assertTrue(r.max() < 0.9)

    def test_ipw_input_assertions(self):
        s_w = np.array((1))
        self.assertRaisesRegex(
            TypeError,
            "must be a pandas DataFrame",
            balance_ipw.ipw,
            s_w,
            s_w,
            s_w,
            s_w,
        )

        s = pd.DataFrame({"a": (1, 2), "id": (1, 2)})
        s_w = pd.Series((1,))
        self.assertRaisesRegex(
            Exception,
            "must be the same length",
            balance_ipw.ipw,
            s,
            s_w,
            s,
            s_w,
        )

    def test_ipw_dropna_empty(self):
        s = pd.DataFrame({"a": (1, None), "b": (np.nan, 2), "id": (1, 2)})
        s_w = pd.Series((1, 2))
        self.assertRaisesRegex(
            Exception,
            "Dropping rows led to empty",
            balance_ipw.ipw,
            s,
            s_w,
            s,
            s_w,
            na_action="drop",
        )

    def test_ipw_formula(self):
        sample = pd.DataFrame(
            {
                "a": (1, 2, 3, 4, 5, 6, 7, 9, 1),
                "b": (1, 2, 3, 4, 5, 6, 7, 9, 1),
                "c": (1, 2, 3, 4, 5, 6, 7, 9, 1),
            }
        )
        target = pd.DataFrame(
            {
                "a": (1, 2, 3, 4, 5, 6, 7, 9, 9),
                "b": (1, 2, 3, 4, 5, 6, 7, 9, 1),
                "c": (1, 2, 3, 4, 5, 6, 7, 9, 1),
            }
        )

        result = balance_ipw.ipw(
            sample_df=sample,
            sample_weights=pd.Series((1,) * 9),
            target_df=target,
            target_weights=pd.Series((1,) * 9),
            formula="a : b + c",
            transformations=None,
        )
        self.assertEqual(result["model"]["X_matrix_columns"], ["a:b", "c"])

    def test_ipw_consistency_with_default_arguments(self):
        # This test is meant to check the consistency of the ipw function with the default arguments,
        # Note that this test rely on all balance functions that are part of ipw:
        # choose_variables, apply_transformations, model_matrix, choose_regularization,
        # weights_from_link, cv_glmnet_performance
        # Therefore a failure in this test may indicate a failure in one
        # of these functions as well as a failure of the ipw function

        np.random.seed(2021)
        n_sample = 1000
        n_target = 2000

        sample_df = pd.concat(
            [
                pd.DataFrame(np.random.uniform(0, 10, size=n_sample), columns=[0]),
                pd.DataFrame(
                    np.random.uniform(0, 1, size=(n_sample, 4)), columns=range(1, 5)
                ),
                pd.DataFrame(
                    np.random.choice(
                        ["level1", "level2", "level3"], size=(n_sample, 5)
                    ),
                    columns=range(5, 10),
                ),
            ],
            axis=1,
        )
        sample_df = sample_df.rename(columns={i: "abcdefghij"[i] for i in range(0, 10)})

        target_df = pd.concat(
            [
                pd.DataFrame(np.random.uniform(8, 18, size=n_target), columns=[0]),
                pd.DataFrame(
                    np.random.uniform(0, 1, size=(n_target, 4)), columns=range(1, 5)
                ),
                pd.DataFrame(
                    np.random.choice(
                        ["level1", "level2", "level3"], size=(n_target, 5)
                    ),
                    columns=range(5, 10),
                ),
            ],
            axis=1,
        )
        target_df = target_df.rename(columns={i: "abcdefghij"[i] for i in range(0, 10)})

        sample_weights = pd.Series(np.random.uniform(0, 1, size=n_sample))
        target_weights = pd.Series(np.random.uniform(0, 1, size=n_target))

        res = balance_ipw.ipw(
            sample_df, sample_weights, target_df, target_weights, max_de=1.5
        )

        # Compare output weights (examples and distribution)
        self.assertEqual(round(res["weight"][15], 4), 0.0886)
        self.assertEqual(round(res["weight"][995], 4), 0.2363)
        self.assertEqual(
            np.around(res["weight"].describe().values, 4),
            np.array([1000, 1.0167, 0.7108, 0.0001, 0.2349, 1.2151, 1.7077, 1.7077]),
        )

        # Compare properties of output model
        self.assertEqual(np.around(res["model"]["fit"]["lambda_1se"], 5), 0.00474)
        self.assertEqual(
            np.around(res["model"]["perf"]["prop_dev_explained"], 5), 0.77107
        )
        self.assertEqual(np.around(res["model"]["perf"]["mean_cv_error"], 5), 0.36291)
        self.assertEqual(np.around(res["model"]["lambda"], 5), 0.00064)
        self.assertEqual(res["model"]["regularisation_perf"]["best"]["trim"], 0.05)

    @patch("balance.weighting_methods.ipw.scipy")
    def test_patch_scipy_random(self, mock_scipy):
        mock_scipy.random = None
        with balance_ipw._patch_scipy_random():
            self.assertEqual(mock_scipy.random, np.random)
        self.assertEqual(mock_scipy.random, None)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import inspect
import logging

from argparse import ArgumentParser, FileType, Namespace

from typing import Dict, List, Optional, Tuple, Type, Union

import balance

import pandas as pd

from balance import __version__  # @manual
from balance.sample_class import Sample as balance_sample_cls  # @manual

logger: logging.Logger = logging.getLogger(__package__)


class BalanceCLI:
    def __init__(self, args) -> None:
        self.args = args

        # Create attributes (to be populated later, which will be used in main)
        (
            self._transformations,
            self._formula,
            self._penalty_factor,
            self._one_hot_encoding,
            self._max_de,
            self._weight_trimming_mean_ratio,
            self._sample_cls,
            self._sample_package_name,
            self._sample_package_version,
        ) = (None, None, None, None, None, None, None, None, None)

    def check_input_columns(self, columns: Union[List[str], pd.Index]) -> None:
        needed_columns = []
        needed_columns.append(self.sample_column())
        needed_columns.append(self.id_column())
        needed_columns.append(self.weight_column())
        needed_columns.extend(self.covariate_columns())
        if self.has_batch_columns():
            needed_columns.extend(self.batch_columns())
        if self.has_keep_columns():
            needed_columns.extend(self.keep_columns())
        if self.has_keep_row_column():
            needed_columns.append(self.keep_row_column())

        for nc in needed_columns:
            assert nc in columns, f"{nc} not in input colums"

    # TODO: decide if to explicitly mention/check here the option of methods or not
    def method(self) -> str:
        return self.args.method

    def sample_column(self) -> str:
        return self.args.sample_column

    def id_column(self) -> str:
        return self.args.id_column

    def weight_column(self) -> str:
        return self.args.weight_column

    def covariate_columns(self) -> str:
        return self.args.covariate_columns.split(",")

    def covariate_columns_for_diagnostics(self) -> List[str]:
        out = self.args.covariate_columns_for_diagnostics
        return None if out is None else out.split(",")

    def rows_to_keep_for_diagnostics(self) -> str:
        return self.args.rows_to_keep_for_diagnostics

    def has_batch_columns(self) -> bool:
        return self.args.batch_columns is not None

    def batch_columns(self) -> str:
        return self.args.batch_columns.split(",")

    def has_keep_columns(self) -> bool:
        return self.args.keep_columns is not None

    def keep_columns(self):  # TODO: figure out how to type hint this one.
        if self.args.keep_columns:
            return self.args.keep_columns.split(",")

    def has_keep_row_column(self) -> bool:
        return self.args.keep_row_column is not None

    def keep_row_column(self) -> str:
        return self.args.keep_row_column

    def max_de(self) -> Optional[float]:
        return self.args.max_de

    def transformations(self) -> Optional[str]:
        if (self.args.transformations is None) or (self.args.transformations == "None"):
            return None
        else:
            return self.args.transformations

    def formula(self) -> Optional[str]:
        return self.args.formula

    def one_hot_encoding(self) -> Optional[bool]:
        return balance.util._true_false_str_to_bool(self.args.one_hot_encoding)

    def standardize_types(self) -> bool:
        return balance.util._true_false_str_to_bool(self.args.standardize_types)

    def weight_trimming_mean_ratio(self) -> float:
        return self.args.weight_trimming_mean_ratio

    def split_sample(self, df: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
        in_sample = df[self.sample_column()] == 1
        sample_df = df[in_sample]
        target_df = df[~in_sample]

        return sample_df, target_df

    def process_batch(
        self,
        batch_df: pd.DataFrame,
        transformations: Union[Dict, str, None] = "default",
        formula=None,
        penalty_factor=None,
        one_hot_encoding: bool = False,
        max_de: Optional[float] = 1.5,
        weight_trimming_mean_ratio: float = 20,
        sample_cls: Type[balance_sample_cls] = balance_sample_cls,
        sample_package_name: str = __package__,
    ) -> Dict[str, pd.DataFrame]:
        # TODO: add unit tests
        sample_df, target_df = self.split_sample(batch_df)

        if sample_df.shape[0] == 0:
            return {
                "adjusted": pd.DataFrame(),
                "diagnostics": pd.DataFrame(
                    {
                        "metric": ("adjustment_failure", "adjustment_failure_reason"),
                        "var": (None, None),
                        "val": (1, "No input data"),
                    }
                ),
            }

        # Stuff everything that is not id, weight, or covariate into outcomes
        outcome_columns = tuple(
            set(batch_df.columns)
            - {self.id_column(), self.weight_column()}
            - set(self.covariate_columns())
        )

        # definitions for diagnostics
        covariate_columns_for_diagnostics = self.covariate_columns_for_diagnostics()
        rows_to_keep_for_diagnostics = self.rows_to_keep_for_diagnostics()

        sample = sample_cls.from_frame(
            sample_df,
            id_column=self.id_column(),
            weight_column=self.weight_column(),
            outcome_columns=outcome_columns,
            check_id_uniqueness=False,
            standardize_types=self.standardize_types(),
        )
        logger.info("%s sample object: %s" % (sample_package_name, str(sample)))

        target = sample_cls.from_frame(
            target_df,
            id_column=self.id_column(),
            weight_column=self.weight_column(),
            outcome_columns=outcome_columns,
            check_id_uniqueness=False,
            standardize_types=self.standardize_types(),
        )
        logger.info("%s target object: %s" % (sample_package_name, str(target)))

        try:
            adjusted = sample.set_target(target).adjust(
                method=self.method(),  # pyre-ignore[6] it gets str, but the function will verify internally if it's the str it should be.
                transformations=transformations,
                formula=formula,
                penalty_factor=penalty_factor,
                one_hot_encoding=one_hot_encoding,
                max_de=max_de,
                weight_trimming_mean_ratio=weight_trimming_mean_ratio,
            )
            logger.info("Succeeded with adjusting sample to target")
            logger.info("%s adjusted object: %s" % (sample_package_name, str(adjusted)))

            logger.info(
                "Condition on which rows to keep for diagnostics: %s "
                % rows_to_keep_for_diagnostics
            )
            logger.info(
                "Names of columns to keep for diagnostics: %s "
                % covariate_columns_for_diagnostics
            )

            diagnostics = adjusted.keep_only_some_rows_columns(
                rows_to_keep=rows_to_keep_for_diagnostics,
                columns_to_keep=covariate_columns_for_diagnostics,
            ).diagnostics()
            logger.info(
                "%s diagnostics object: %s" % (sample_package_name, str(diagnostics))
            )

            # Update dtypes
            if self.args.return_df_with_original_dtypes:
                df_to_return = balance.util._astype_in_df_from_dtypes(
                    adjusted.df, adjusted._df_dtypes
                )
                balance.util._warn_of_df_dtypes_change(
                    adjusted.df.dtypes,
                    df_to_return.dtypes,
                    "df_after_adjust",
                    "df_returned_by_the_cli",
                )

            else:
                df_to_return = adjusted.df
            rval = {"adjusted": df_to_return, "diagnostics": diagnostics}
        except Exception as e:
            if self.args.succeed_on_weighting_failure:
                logger.error(
                    "Adjustment failed. Because '--succeed_on_weighting_failure' was set: returning empty weights."
                )
                sample.set_weights(None)
                module = inspect.getmodule(inspect.trace()[-1][0])
                module_name = module.__name__ if module is not None else None
                error_message = f"{module_name}: {e}"
                logger.exception("The error message is: " + error_message)
                # Update dtypes
                if self.args.return_df_with_original_dtypes:
                    df_to_return = balance.util._astype_in_df_from_dtypes(
                        sample.df, sample._df_dtypes
                    )
                    balance.util._warn_of_df_dtypes_change(
                        sample.df.dtypes,
                        df_to_return.dtypes,
                        "df_without_adjusting",
                        "df_returned_by_the_cli",
                    )
                else:
                    df_to_return = sample.df
                rval = {
                    "adjusted": df_to_return,
                    "diagnostics": pd.DataFrame(
                        {
                            "metric": (
                                "adjustment_failure",
                                "adjustment_failure_reason",
                            ),
                            "var": (None, None),
                            "val": (1, error_message),
                        }
                    ),
                }
            else:
                raise e
        return rval

    def adapt_output(self, output_df: pd.DataFrame) -> pd.DataFrame:
        """Filter raw output dataframe to user's requested rows/columns.
        - First we filter to the rows we are supposed to keep.
        - Next we select the columns that need to be returned.
        """
        if output_df.empty:
            return output_df

        if self.has_keep_row_column():
            keep_rows = output_df[self.keep_row_column()] == 1
            output_df = output_df[keep_rows]

        if self.has_keep_columns():
            output_df = output_df[  # pyre-ignore[9]: this uses the DataFrame also.
                self.keep_columns()
            ]

        return output_df

    def load_and_check_input(self) -> pd.DataFrame:
        # TODO: Add unit tests for function
        # Load and check input
        input_df = pd.read_csv(self.args.input_file, sep=self.args.sep_input_file)
        logger.info("Number of rows in input file: %d" % input_df.shape[0])
        if self.has_keep_row_column():
            logger.info(
                "Number of rows to keep in input file: %d"
                % input_df[input_df[self.keep_row_column()] == 1].shape[0]
            )
        logger.info("Number of columns in input file: %d" % input_df.shape[1])
        return input_df

    def write_outputs(self, output_df, diagnostics_df) -> None:
        # TODO: Add unit tests for function
        # Write output
        output_df.to_csv(
            path_or_buf=self.args.output_file,
            index=False,
            header=(not self.args.no_output_header),
            sep=self.args.sep_output_file,
        )
        self.args.output_file.close()

        if self.args.diagnostics_output_file is not None:
            diagnostics_df.to_csv(
                path_or_buf=self.args.diagnostics_output_file,
                index=False,
                header=(not self.args.no_output_header),
                sep=self.args.sep_diagnostics_output_file,
            )
            self.args.diagnostics_output_file.close()

    def update_attributes_for_main_used_by_adjust(self) -> None:
        """
        Prepares all the defaults for main to use.
        """
        # TODO: future version might include conditional control over these attributes based on some input
        transformations = self.transformations()
        formula = self.formula()
        penalty_factor = None
        one_hot_encoding = self.one_hot_encoding()
        max_de = self.max_de()
        weight_trimming_mean_ratio = self.weight_trimming_mean_ratio()
        sample_cls, sample_package_name, sample_package_version = (
            balance_sample_cls,
            __package__,
            __version__,
        )

        # update all attributes (to be later used in main)
        (
            self._transformations,
            self._formula,
            self._penalty_factor,
            self._one_hot_encoding,
            self._max_de,
            self._weight_trimming_mean_ratio,
            self._sample_cls,
            self._sample_package_name,
            self._sample_package_version,
        ) = (
            transformations,
            formula,
            penalty_factor,
            one_hot_encoding,
            max_de,
            weight_trimming_mean_ratio,
            sample_cls,
            sample_package_name,
            sample_package_version,
        )

    def main(self) -> None:
        # update all the objects from self attributes
        (
            transformations,
            formula,
            penalty_factor,
            one_hot_encoding,
            max_de,
            weight_trimming_mean_ratio,
            sample_cls,
            sample_package_name,
            sample_package_version,
        ) = (
            self._transformations,
            self._formula,
            self._penalty_factor,
            self._one_hot_encoding,
            self._max_de,
            self._weight_trimming_mean_ratio,
            self._sample_cls,
            self._sample_package_name,
            self._sample_package_version,
        )

        logger.info(
            "Running cli.main() using %s version %s"
            % (sample_package_name, sample_package_version)
        )

        # Logging arguments used by main:
        keys = (
            "transformations",
            "formula",
            "penalty_factor",
            "one_hot_encoding",
            "max_de",
            "weight_trimming_mean_ratio",
            "sample_cls",
            "sample_package_name",
            "sample_package_version",
        )
        values = (
            transformations,
            formula,
            penalty_factor,
            one_hot_encoding,
            max_de,
            weight_trimming_mean_ratio,
            sample_cls,
            sample_package_name,
            sample_package_version,
        )
        main_config = dict(zip(keys, values))
        logger.info("Attributes used by main() for running adjust: %s" % main_config)

        # Load and check input
        input_df = self.load_and_check_input()
        self.check_input_columns(input_df.columns)

        # Run process_batch on input_df, and adjustment arguments
        if self.has_batch_columns():
            results = []
            diagnostics = []
            for batch_name, batch_df in input_df.groupby(self.batch_columns()):
                logger.info("Running weighting for batch = %s " % str(batch_name))
                processed = self.process_batch(
                    batch_df,
                    transformations,
                    formula,
                    penalty_factor,
                    one_hot_encoding,
                    max_de,
                    weight_trimming_mean_ratio,
                    sample_cls,
                    sample_package_name,
                )
                results.append(processed["adjusted"])
                diagnostics.append(processed["diagnostics"])
                logger.info("Done processing batch %s" % str(batch_name))

            if (len(results) == 0) and len(diagnostics) == 0:
                output_df = pd.DataFrame()
                diagnostics_df = pd.DataFrame()
            else:
                output_df = pd.concat(results)
                diagnostics_df = pd.concat(diagnostics)
        else:
            processed = self.process_batch(
                input_df,
                transformations,
                formula,
                penalty_factor,
                one_hot_encoding,
                max_de,
                weight_trimming_mean_ratio,
                sample_cls,
                sample_package_name,
            )
            output_df = processed["adjusted"]
            diagnostics_df = processed["diagnostics"]

        logger.info("Done fitting the model, writing output")
        # Remove unneeded rows and columns
        output_df = self.adapt_output(output_df)

        # Write output
        self.write_outputs(output_df, diagnostics_df)

    def __del__(self) -> None:
        for handle in [self.args.input_file, self.args.output_file]:
            try:
                handle.close()
            except FileNotFoundError:
                pass


def _float_or_none(value: Union[float, int, str, None]) -> Optional[float]:
    """Return a float (if float or int) or None if it's None or "None"

    This is so as to be clear that some input returned type is float or None.

    Args:
        value (Union[float, int, str, None]): A value to be converted.

    Returns:
        Optional[float]: None or float.
    """
    if (value is None) or (value == "None"):
        return None
    return float(value)


def add_arguments_to_parser(parser: ArgumentParser) -> ArgumentParser:
    # TODO: add checks for validity of input (including None as input)
    # TODO: add arguments for formula when used as a list and for penalty_factor
    parser.add_argument(
        "--input_file",
        type=FileType("r"),
        required=True,
        help="Path to input sample/target",
    )
    parser.add_argument(
        "--output_file",
        type=FileType("w"),
        required=True,
        help="Path to write output weights",
    )
    parser.add_argument(
        "--diagnostics_output_file",
        type=FileType("w"),
        required=False,
        help="Path to write adjustment diagnostics",
    )
    parser.add_argument(
        "--method", default="ipw", help="Method to use for weighting [default=ipw]"
    )
    parser.add_argument(
        "--sample_column",
        default="is_respondent",
        help="Path to target population [default=is_respondent]",
    )
    parser.add_argument(
        "--id_column",
        default="id",
        help="Column that identifies units [default=id]",
    )
    parser.add_argument(
        "--weight_column",
        default="weight",
        help="Column that identifies weights of samples [default=weight]",
    )
    parser.add_argument(
        "--covariate_columns", required=True, help="Set of columns used for adjustment"
    )
    parser.add_argument(
        "--covariate_columns_for_diagnostics",
        required=False,
        default=None,
        help="Set of columns used for diagnostics reporting (if not supplied the default is None, which means to use all columns from --covariate_columns)",
    )
    parser.add_argument(
        "--rows_to_keep_for_diagnostics",
        required=False,
        default=None,
        help="A string with a condition for filtering rows (to be used for diagnostics reporting). \
             The condition should be based on the list of covariate_columns and result in a bool pd.Series \
             (e.g.: 'is_married' or 'is_married & age >= 18', etc.) \
             (if not supplied the default is None, which means to use all rows without filtering",
    )
    parser.add_argument(
        "--batch_columns",
        required=False,
        help="Set of columns used to indicate batches of data",
    )
    parser.add_argument(
        "--keep_columns",
        type=str,
        required=False,
        help="Set of columns we include in the output csv file",
    )
    parser.add_argument(
        "--keep_row_column",
        type=str,
        required=False,
        help="Column indicating which rows we include in the output csv file",
    )
    parser.add_argument(
        "--sep_input_file",
        type=str,
        required=False,
        default=",",
        help="A 1 character for indicating the delimiter for the output file. If not supplied it defaults to a comma (,)",
    )
    parser.add_argument(
        "--sep_output_file",
        type=str,
        required=False,
        default=",",
        help="A 1 character for indicating the delimiter for the output file. If not supplied it defaults to a comma (,)",
    )
    parser.add_argument(
        "--sep_diagnostics_output_file",
        type=str,
        required=False,
        default=",",
        help="A 1 character for indicating the delimiter for the diagnostics output file. If not supplied it defaults to a comma (,)",
    )
    parser.add_argument(
        "--no_output_header",
        default=False,
        action="store_true",
        help="Turn off header in the output csv file",
    )
    parser.add_argument(
        "--succeed_on_weighting_failure",
        action="store_true",
        help=(
            "If adjustment fails (e.g. because the input data has too few "
            "rows), then do not fail, but instead return the input data with "
            "all weights null"
        ),
    )
    parser.add_argument(
        "--max_de",
        type=_float_or_none,
        required=False,
        default=1.5,
        help=(
            "Upper bound for the design effect of the computed weights. "
            "If not supplied it defaults to 1.5. If set to 'None', then it uses 'lambda_1se'. "
            "Only used if method is ipw or CBPS."
        ),
    )
    parser.add_argument(
        "--weight_trimming_mean_ratio",
        type=_float_or_none,
        required=False,
        default=20.0,
        help=(
            "The ratio according to which weights are trimmed from above by mean(weights) * ratio. "
            "Defaults to 20. "
            "Used only for CBPS and ipw."
            "For ipw this is only used if max_de is set to 'None',otherwise, the trimming ratio is chosen by the algorithm."
        ),
    )
    parser.add_argument(
        "--one_hot_encoding",
        type=str,
        default="True",
        required=False,
        help=(
            "Set the value of the one_hot_encoding parameter. Accepts a string with one a value of 'True' or 'False' (treats it as a bool). Default is 'True'"
        ),
    )
    # TODO: Ideally we would like transformations argument to be able to get three types of values: None (for no transformations),
    # "default" for default transformations or a dictionary of transformations.
    # However, as a first step I added the option for "default" (which is also the default) and None (for no transformations).
    parser.add_argument(
        "--transformations",
        default="default",
        required=False,
        help=(
            "Define the transformations for the covariates. Can be set to None for no transformations or"
            "'default' for default transformations."
        ),
    ),
    # TODO: we currently support only the option of a string formula (or None), not a list of formulas.
    parser.add_argument(
        "--formula",
        default=None,
        required=False,
        help=(
            "The formula of the model matrix (in ipw or cbps). If None (default), the formula will be setted to an additive formula using all the covariates."
        ),
    )
    parser.add_argument(
        "--return_df_with_original_dtypes",
        action="store_true",
        help=(
            "If the input table has unsupported column types (e.g.: int32), then when using Sample.from_frame it will be changed (e.g.: float32). "
            "The returned df from the cli will use the transformed DataFrame. If this flag is used, then the cli will attempt to restore the original dtypes of the input table"
            "before returning it."
            "WARNING: sometimes the pd.astype command might lead to odd behaviors, so it is generally safer to NOT use this flag but to manually specify the desired dtypes in the input/output tables."
            "For example, dealing with missing values could lead to many issues (e.g.: there is np.nan and pd.NA, and these do not play nicely with type conversions)"
        ),
    )
    parser.add_argument(
        "--standardize_types",
        type=str,
        default="True",
        required=False,
        help=(
            "Control the standardize_types argument in Sample.from_frame (which is used on the files read by the cli)"
            "The default is True. It is generally not advised to use False since this step is needed to deal with converting input types that are not supported by various functions."
            "For example, if a column of Int64 has pandas.NA, it could fail on various functions. Current default (True) will turn that column into float64"
            "(the pandas.NA will be converted into numpy.nan)."
            "Setting the current flag to 'False' might lead to failures. The advantage is that it would keep most columns dtype as is"
            " (which might be helpful for some downstream operations that assume the output dtypes are the same as the input dtypes)."
            "NOTE: regardless if this flag is set to true or false, the weight column will be turned into a float64 type anyway."
        ),
    )

    return parser


def make_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser = add_arguments_to_parser(parser)
    return parser


if __name__ == "__main__":
    parser: ArgumentParser = make_parser()
    args: Namespace = parser.parse_args()
    cli = BalanceCLI(args)
    cli.update_attributes_for_main_used_by_adjust()
    cli.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import pandas as pd

logger: logging.Logger = logging.getLogger(__package__)


##########################################
# Weights diagnostics - functions for analyzing weights
##########################################


def _check_weights_are_valid(
    w: Union[
        List,
        pd.Series,
        np.ndarray,
        pd.DataFrame,
        None,
    ]
) -> None:
    """Check weights.

    Args:
        w (Union[ List, pd.Series, np.ndarray, pd.DataFrame, None, ]): input weights.
            If w is pd.DataFrame then only the first column will be checked (assuming it is a column of weights).
            If input is None, then the function returns None with no errors (since None is a valid weights input for various functions).

    Raises:
        ValueError: if weights are not numeric.
        ValueError: if weights include a negative value.

    Returns:
        _type_: None
    """
    if w is None:
        return None
    if isinstance(w, pd.DataFrame):
        w = w.iloc[:, 0]  # if DataFrame, we check only the first column.
    if not isinstance(w, pd.Series):
        w = pd.Series(w)
        # TODO: (p2) consider having a check for each type of w, instead of
        #            turning w into pd.Series (since this solution might not be very efficient)
    if not np.issubdtype(w, np.number):
        raise TypeError(
            f"weights (w) must be a number but instead they are of type: {w.dtype}."
        )
    if any(w < 0):
        raise ValueError("weights (w) must all be non-negative values.")
    # TODO: do we also want to verify that at least one weight is larger than 0?!

    return None


# TODO: if the input is pd.DataFrame than the output will be pd.Series.
#       we could make the support of this more official in the future.
def design_effect(w: pd.Series) -> np.float64:
    """
    Kish's design effect measure.

    For details, see:
    - https://en.wikipedia.org/wiki/Design_effect
    - https://en.wikipedia.org/wiki/Effective_sample_size

    Args:
        w (pd.Series): A pandas series of weights (non negative, float/int) values.

    Returns:
        np.float64: An estimator saying by how much the variance of the mean is expected to increase, compared to a random sample mean,
        due to application of the weights.

    Examples:
        ::

            from balance.stats_and_plots.weights_stats import design_effect
            import pandas as pd

            design_effect(pd.Series((0, 1, 2, 3)))
                # output:
                # 1.5555555555555556
            design_effect(pd.Series((1, 1, 1000)))
                # 2.9880418803112336
                # As expected. With a single dominating weight - the Deff is almost equal to the sample size.
    """
    _check_weights_are_valid(w)
    return (w**2).mean() / (w.mean() ** 2)


def nonparametric_skew(w: pd.Series) -> np.float64:
    # TODO (p2): consider adding other skew measures (https://en.wikipedia.org/wiki/Skewness)
    #            look more in the literature (are there references for using this vs another, or none at all?)
    #            update the doc with insights, once done:
    #            what is more accepted in the literature in the field and what are the advantages of each.
    #            Any reference to literature where this is used to analyze weights of survey?
    #            Add reference to some interpretation of these values?
    """
    The nonparametric skew is the difference between the mean and the median, divided by the standard deviation.
    See:
    - https://en.wikipedia.org/wiki/Nonparametric_skew

    Args:
        w (pd.Series): A pandas series of weights (non negative, float/int) values.

    Returns:
        np.float64: A value of skew, between -1 to 1, but for weights it's often positive (i.e.: right tailed distribution).
        The value returned will be 0 if the standard deviation is 0 (i.e.: all values are identical), or if the input is of length 1.

    Examples:
        ::

            from balance.stats_and_plots.weights_stats import nonparametric_skew

            nonparametric_skew(pd.Series((1, 1, 1, 1)))  # 0
            nonparametric_skew(pd.Series((1)))           # 0
            nonparametric_skew(pd.Series((1, 2, 3, 4)))  # 0
            nonparametric_skew(pd.Series((1, 1, 1, 2)))  # 0.5
            nonparametric_skew(pd.Series((-1,1,1, 1)))   #-0.5

    """
    _check_weights_are_valid(w)
    if (len(w) == 1) or (w.std() == 0):
        return np.float64(0)
    return (w.mean() - w.median()) / w.std()


def prop_above_and_below(
    w: pd.Series,
    below: Union[Tuple[float, ...], List[float], None] = (
        1 / 10,
        1 / 5,
        1 / 3,
        1 / 2,
        1,
    ),
    above: Union[Tuple[float, ...], List[float], None] = (1, 2, 3, 5, 10),
    return_as_series: bool = True,
) -> Union[pd.Series, Dict[Any, Any], None]:
    # TODO (p2): look more in the literature (are there references for using this vs another, or none at all?)
    #            update the doc with insights, once done.
    """
    The proportion of weights, normalized to sample size, that are above and below some numbers (E.g. 1,2,3,5,10 and their inverse: 1, 1/2, 1/3, etc.).
    This is similar to returning percentiles of the (normalized) weighted distribution. But instead of focusing on the 25th percentile, the median, etc,
    We focus instead on more easily interpretable weights values.

    For example, saying that some proportion of users had a weight of above 1 gives us an indication of how many users
    we got that we don't "loose" their value after using the weights. Saying which proportion of users had a weight below 1/10 tells us how many users
    had basically almost no contribution to the final analysis (after applying the weights).

    Note that below and above can overlap, be unordered, etc. The user is responsible for the order.

    Args:
        w (pd.Series): A pandas series of weights (float, non negative) values.
        below (Union[Tuple[float, ...], List[float], None], optional):
            values to check which proportion of normalized weights are *below* them.
            Using None returns None.
            Defaults to (1/10, 1/5, 1/3, 1/2, 1).
        above (Union[Tuple[float, ...], List[float], None], optional):
            values to check which proportion of normalized weights are *above* (or equal) to them.
            Using None returns None.
            Defaults to (1, 2, 3, 5, 10).
        return_as_series (bool, optional): If true returns one pd.Series of values.
            If False will return a dict with two pd.Series (one for below and one for above).
            Defaults to True.

    Returns:
        Union[pd.Series, Dict]:
        If return_as_series is True we get pd.Series with proportions of (normalized weights)
        that are below/above some numbers, the index indicates which threshold was checked
        (the values in the index are rounded up to 3 points for printing purposes).
        If return_as_series is False we get a dict with 'below' and 'above' with the relevant pd.Series (or None).

    Examples:
        ::

            from balance.stats_and_plots.weights_stats import prop_above_and_below
            import pandas as pd

            # normalized weights:
            print(pd.Series((1, 2, 3, 4)) / pd.Series((1, 2, 3, 4)).mean())
                # 0    0.4
                # 1    0.8
                # 2    1.2
                # 3    1.6

            # checking the function:
            prop_above_and_below(pd.Series((1, 2, 3, 4)))
                # dtype: float64
                # prop(w < 0.1)      0.00
                # prop(w < 0.2)      0.00
                # prop(w < 0.333)    0.00
                # prop(w < 0.5)      0.25
                # prop(w < 1.0)      0.50
                # prop(w >= 1)       0.50
                # prop(w >= 2)       0.00
                # prop(w >= 3)       0.00
                # prop(w >= 5)       0.00
                # prop(w >= 10)      0.00
                # dtype: float64

            prop_above_and_below(pd.Series((1, 2, 3, 4)), below = (0.1, 0.5), above = (2,3))
                # prop(w < 0.1)    0.00
                # prop(w < 0.5)    0.25
                # prop(w >= 2)     0.00
                # prop(w >= 3)     0.00
                # dtype: float64

            prop_above_and_below(pd.Series((1, 2, 3, 4)), return_as_series = False)
                # {'below': prop(w < 0.1)      0.00
                # prop(w < 0.2)      0.00
                # prop(w < 0.333)    0.00
                # prop(w < 0.5)      0.25
                # prop(w < 1)        0.50
                # dtype: float64, 'above': prop(w >= 1)     0.5
                # prop(w >= 2)     0.0
                # prop(w >= 3)     0.0
                # prop(w >= 5)     0.0
                # prop(w >= 10)    0.0
                # dtype: float64}

    """
    _check_weights_are_valid(w)

    # normalize weight to sample size:
    w = w / w.mean()

    if below is None and above is None:
        return None

    # calculate props from below:
    if below is not None:
        prop_below = [(w < i).mean() for i in below]
        prop_below_index = ["prop(w < " + str(round(i, 3)) + ")" for i in below]
        prop_below_series = pd.Series(prop_below, index=prop_below_index)
    else:
        prop_below_series = None

    # calculate props from above:
    if above is not None:
        prop_above = [(w >= i).mean() for i in above]
        prop_above_index = ["prop(w >= " + str(round(i, 3)) + ")" for i in above]
        prop_above_series = pd.Series(prop_above, index=prop_above_index)
    else:
        prop_above_series = None

    # decide if to return one series or a dict
    if return_as_series:
        out = pd.concat(
            [  # pyre-ignore[6]: pd.concat supports Series.
                prop_below_series,
                prop_above_series,
            ]
        )
    else:
        out = {"below": prop_below_series, "above": prop_above_series}

    return out  # pyre-ignore[7]:  TODO: see if we can fix this pyre


def weighted_median_breakdown_point(w: pd.Series) -> np.float64:
    # TODO (p2): do we want to have weighted_quantile_breakdown_point
    # so to check for quantiles other than 50%?
    """
    Calculates the minimal percent of users that have at least 50% of the weights.
    This gives us the breakdown point of calculating the weighted median.
    This can be thought of as reflecting a similar metric to the design effect.
    See also:
    - https://en.wikipedia.org/wiki/Weighted_median
    - https://en.wikipedia.org/wiki/Robust_statistics#Breakdown_point

    Args:
        w (pd.Series): A pandas series of weights (float, non negative values).

    Returns:
        np.float64: A minimal percent of users that contain at least 50% of the weights.

    Examples:
        ::

            w = pd.Series([1,1,1,1])
            print(weighted_median_breakdown_point(w)) # 0.5

            w = pd.Series([2,2,2,2])
            print(weighted_median_breakdown_point(w)) # 0.5

            w = pd.Series([1,1,1, 10])
            print(weighted_median_breakdown_point(w)) # 0.25

            w = pd.Series([1,1,1,1, 10])
            print(weighted_median_breakdown_point(w)) # 0.2
    """
    _check_weights_are_valid(w)

    # normalize weight to sample size:

    n = len(w)  # n users
    w = w / w.sum()  # normalize to 1
    # get a cumsum of sorted weights to find the median:
    w_freq_cumsum = w.sort_values(  # pyre-ignore[16]: it does have a cumsum method.
        ascending=False
    ).cumsum()
    numerator = (w_freq_cumsum <= 0.5).sum()
    if numerator == 0:
        numerator = (
            1  # this happens if one observation has more than 50% of the weights
        )
    # find minimal proportion of samples needed to reach 50%
    # the +1 trick is to deal with cases that 1 user has a weight that is larget then 50%.
    return numerator / n  # breakdown_point



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from copy import deepcopy

import balance.testutil

import numpy as np
import numpy.testing
import pandas as pd

# TODO: remove the use of balance_util in most cases, and just import the functions to be tested directly
from balance import util as balance_util
from balance.sample_class import Sample

from numpy import dtype

from patsy import dmatrix  # pyre-ignore[21]: this module exists
from scipy.sparse import csc_matrix


class TestUtil(
    balance.testutil.BalanceTestCase,
):
    def test__check_weighting_methods_input(self):
        self.assertRaisesRegex(
            TypeError,
            "sample_df must be a pandas DataFrame",
            balance_util._check_weighting_methods_input,
            df=1,
            weights=pd.Series(1),
            object_name="sample",
        )
        self.assertRaisesRegex(
            TypeError,
            "sample_weights must be a pandas Series",
            balance_util._check_weighting_methods_input,
            df=pd.DataFrame({"a": [1]}),
            weights="a",
            object_name="sample",
        )

        self.assertRaisesRegex(
            ValueError,
            "sample_weights must be the same length as sample_df",
            balance_util._check_weighting_methods_input,
            df=pd.DataFrame({"a": [1, 2]}),
            weights=pd.Series([1]),
            object_name="sample",
        )

        self.assertRaisesRegex(
            ValueError,
            "sample_df index must be the same as sample_weights index",
            balance_util._check_weighting_methods_input,
            df=pd.DataFrame({"a": [1, 2]}, index=[1, 2]),
            weights=pd.Series([1, 1], index=[3, 4]),
            object_name="sample",
        )

    def test_guess_id_column(self):
        # test when id column is presented
        df = pd.DataFrame(
            {
                "a": (0, 1, 2),
                "id": (1, 2, 3),
            }
        )
        self.assertEqual(balance_util.guess_id_column(df), "id")
        self.assertWarnsRegexp(
            "Guessed id column name id for the data",
            balance_util.guess_id_column,
            df,
        )

        # test when column_name is passed
        df = pd.DataFrame(
            {
                "a": (0, 1, 2),
                "b": (1, 2, 3),
            }
        )
        self.assertEqual(balance_util.guess_id_column(df, column_name="b"), "b")
        with self.assertRaisesRegex(
            ValueError,
            "Dataframe does not have column*",
        ):
            balance_util.guess_id_column(df, column_name="c")

        # test when no id column is passed and no id column in dataframe
        with self.assertRaisesRegex(
            ValueError,
            "Cannot guess id column name for this DataFrame. Please provide a value in id_column",
        ):
            balance_util.guess_id_column(df)

    def test__isinstance_sample(self):
        from balance.util import _isinstance_sample

        s_df = pd.DataFrame(
            {
                "a": (0, 1, 2),
                "b": (0, None, 2),
                "c": ("a", "b", "a"),
                "id": (1, 2, 3),
            }
        )
        s = Sample.from_frame(s_df)
        self.assertFalse(_isinstance_sample(s_df))
        self.assertTrue(_isinstance_sample(s))

    def test_add_na_indicator(self):
        df = pd.DataFrame({"a": (0, None, 2, np.nan), "b": (None, "b", "", np.nan)})
        e = pd.DataFrame(
            {
                "a": (0, 0, 2.0, 0),
                "b": ("_NA", "b", "", "_NA"),
                "_is_na_a": (False, True, False, True),
                "_is_na_b": (True, False, False, True),
            },
            columns=("a", "b", "_is_na_a", "_is_na_b"),
        )
        r = balance_util.add_na_indicator(df)
        self.assertEqual(r, e)

        # No change if no missing variables
        df = pd.DataFrame(
            {"a": (0, 1, 2), "b": ("a", "b", ""), "c": pd.Categorical(("a", "b", "a"))}
        )
        self.assertEqual(balance_util.add_na_indicator(df), df)

        # Test that it works with categorical variables
        df = pd.DataFrame(
            {
                "c": pd.Categorical(("a", "b", "a", "b")),
                "d": pd.Categorical(("a", "b", None, np.nan)),
            }
        )
        e = pd.DataFrame(
            {
                "c": pd.Categorical(("a", "b", "a", "b")),
                "d": pd.Categorical(
                    ("a", "b", "_NA", "_NA"), categories=("a", "b", "_NA")
                ),
                "_is_na_d": (False, False, True, True),
            },
            columns=("c", "d", "_is_na_d"),
        )
        self.assertEqual(balance_util.add_na_indicator(df), e)

        # test arguments
        df = pd.DataFrame({"a": (0, None, 2, np.nan), "b": (None, "b", "", np.nan)})
        e = pd.DataFrame(
            {
                "a": (0.0, 42.0, 2.0, 42.0),
                "b": ("AAA", "b", "", "AAA"),
                "_is_na_a": (False, True, False, True),
                "_is_na_b": (True, False, False, True),
            },
            columns=("a", "b", "_is_na_a", "_is_na_b"),
        )
        r = balance_util.add_na_indicator(df, replace_val_obj="AAA", replace_val_num=42)
        self.assertEqual(r, e)

        # check exceptions
        d = pd.DataFrame({"a": [0, 1, np.nan, None], "b": ["x", "y", "_NA", None]})
        self.assertRaisesRegex(
            Exception,
            "Can't add NA indicator to columns containing NAs and the value '_NA', ",
            balance_util.add_na_indicator,
            d,
        )
        d = pd.DataFrame({"a": [0, 1, np.nan, None], "_is_na_b": ["x", "y", "z", None]})
        self.assertRaisesRegex(
            Exception,
            "Can't add NA indicator to DataFrame which contains",
            balance_util.add_na_indicator,
            d,
        )

    def test_drop_na_rows(self):
        sample_df = pd.DataFrame(
            {"a": (0, None, 2, np.nan), "b": (None, "b", "c", np.nan)}
        )
        sample_weights = pd.Series([1, 2, 3, 4])
        (
            sample_df,
            sample_weights,
        ) = balance_util.drop_na_rows(sample_df, sample_weights, "sample")
        self.assertEqual(sample_df, pd.DataFrame({"a": (2.0), "b": ("c")}, index=[2]))
        self.assertEqual(sample_weights, pd.Series([3], index=[2]))

        # check exceptions
        sample_df = pd.DataFrame({"a": (None), "b": ("b")}, index=[1])
        sample_weights = pd.Series([1])
        self.assertRaisesRegex(
            ValueError,
            "Dropping rows led to empty",
            balance_util.drop_na_rows,
            sample_df,
            sample_weights,
            "sample",
        )

    def test_formula_generator(self):
        self.assertEqual(balance_util.formula_generator("a"), "a")
        self.assertEqual(balance_util.formula_generator(["a", "b", "c"]), "c + b + a")
        # check exceptions
        self.assertRaisesRegex(
            Exception,
            "This formula type is not supported",
            balance_util.formula_generator,
            ["a", "b"],
            "interaction",
        )

    def test_dot_expansion(self):
        self.assertEqual(
            balance_util.dot_expansion(".", ["a", "b", "c", "d"]), "(a+b+c+d)"
        )
        self.assertEqual(
            balance_util.dot_expansion("b:(. - a)", ["a", "b", "c", "d"]),
            "b:((a+b+c+d) - a)",
        )
        self.assertEqual(balance_util.dot_expansion("a*b", ["a", "b", "c", "d"]), "a*b")
        d = {"a": ["a1", "a2", "a1", "a1"]}
        df = pd.DataFrame(data=d)
        self.assertEqual(balance_util.dot_expansion(".", list(df.columns)), "(a)")

        # check exceptions
        self.assertRaisesRegex(
            Exception,
            "Variables should not be empty. Please provide a list of strings.",
            balance_util.dot_expansion,
            ".",
            None,
        )
        self.assertRaisesRegex(
            Exception,
            "Variables should be a list of strings and have to be included.",
            balance_util.dot_expansion,
            ".",
            df,
        )

    def test_process_formula(self):
        from patsy import EvalFactor, Term  # pyre-ignore[21]: this module exists

        f1 = balance_util.process_formula("a:(b+aab)", ["a", "b", "aab"])
        self.assertEqual(
            f1.rhs_termlist,
            [
                Term([EvalFactor("a"), EvalFactor("b")]),
                Term([EvalFactor("a"), EvalFactor("aab")]),
            ],
        )

        f2 = balance_util.process_formula("a:(b+aab)", ["a", "b", "aab"], ["a", "b"])
        self.assertEqual(
            f2.rhs_termlist,
            [
                Term(
                    [
                        EvalFactor("C(a, one_hot_encoding_greater_2)"),
                        EvalFactor("C(b, one_hot_encoding_greater_2)"),
                    ]
                ),
                Term(
                    [EvalFactor("C(a, one_hot_encoding_greater_2)"), EvalFactor("aab")]
                ),
            ],
        )

        # check exceptions
        self.assertRaisesRegex(
            Exception,
            "Not all factor variables are contained in variables",
            balance_util.process_formula,
            formula="a:(b+aab)",
            variables=["a", "b", "aab"],
            factor_variables="c",
        )

    def test_build_model_matrix(self):
        df = pd.DataFrame(
            {"a": ["a1", "a2", "a1", "a1"], "b": ["b1", "b2", "b3", "b3"]}
        )
        res = pd.DataFrame(
            {"a[a1]": (1.0, 0.0, 1.0, 1.0), "a[a2]": (0.0, 1.0, 0.0, 0.0)}
        )
        # explicit formula
        x_matrix = balance_util.build_model_matrix(df, "a")
        self.assertEqual(x_matrix["model_matrix"], res)
        self.assertEqual(x_matrix["model_matrix_columns"], res.columns.tolist())

        # formula with dot
        x_matrix = balance_util.build_model_matrix(df, ".")
        res = pd.DataFrame(
            {
                "a[a1]": (1.0, 0.0, 1.0, 1.0),
                "a[a2]": (0.0, 1.0, 0.0, 0.0),
                "b[T.b2]": (0.0, 1.0, 0.0, 0.0),
                "b[T.b3]": (0.0, 0.0, 1.0, 1.0),
            }
        )
        self.assertEqual(x_matrix["model_matrix"], res)
        self.assertEqual(x_matrix["model_matrix_columns"], res.columns.tolist())

        # formula with factor_variables
        x_matrix = balance_util.build_model_matrix(df, ".", factor_variables=["a"])
        res = pd.DataFrame(
            {
                "C(a, one_hot_encoding_greater_2)[a2]": (0.0, 1.0, 0.0, 0.0),
                "b[T.b2]": (0.0, 1.0, 0.0, 0.0),
                "b[T.b3]": (0.0, 0.0, 1.0, 1.0),
            }
        )
        self.assertEqual(x_matrix["model_matrix"], res)
        self.assertEqual(x_matrix["model_matrix_columns"], res.columns.tolist())

        # Sparse output
        x_matrix = balance_util.build_model_matrix(df, "a", return_sparse=True)
        res = [[1.0, 0.0], [0.0, 1.0], [1.0, 0.0], [1.0, 0.0]]
        self.assertEqual(x_matrix["model_matrix"].toarray(), res)
        self.assertEqual(x_matrix["model_matrix_columns"], ["a[a1]", "a[a2]"])
        self.assertTrue(type(x_matrix["model_matrix"]) is csc_matrix)

        # Check exceptions
        self.assertRaisesRegex(
            Exception,
            "Not all factor variables are contained in df",
            balance_util.build_model_matrix,
            df,
            formula="a",
            factor_variables="c",
        )

        df = pd.DataFrame({"[a]": ["a1", "a2", "a1", "a1"]})
        self.assertRaisesRegex(
            Exception,
            "Variable names cannot contain characters",
            balance_util.build_model_matrix,
            df,
            "a",
        )

        # Int64Dtype input
        df = pd.DataFrame({"a": [1, 2, 3, 4]})
        df = df.astype(dtype={"a": "Int64"})
        res = pd.DataFrame({"a": (1.0, 2.0, 3.0, 4.0)})
        # explicit formula
        x_matrix = balance_util.build_model_matrix(df, "a")
        self.assertEqual(x_matrix["model_matrix"], res)
        self.assertEqual(x_matrix["model_matrix_columns"], res.columns.tolist())

    def test_model_matrix(self):
        s_df = pd.DataFrame(
            {
                "a": (0, 1, 2),
                "b": (0, None, 2),
                "c": ("a", "b", "a"),
                "id": (1, 2, 3),
            }
        )
        s = Sample.from_frame(s_df)

        # Tests on a single sample
        e = pd.DataFrame(
            {
                "a": (0.0, 1.0, 2.0),
                "b": (0.0, 0.0, 2.0),
                "_is_na_b[T.True]": (0.0, 1.0, 0.0),
                "c[a]": (1.0, 0.0, 1.0),
                "c[b]": (0.0, 1.0, 0.0),
            }
        )
        r = balance_util.model_matrix(s)
        self.assertEqual(r["sample"], e, lazy=True)
        self.assertTrue(r["target"] is None)

        # Tests on a single sample dataframe
        e = pd.DataFrame(
            {
                "a": (0.0, 1.0, 2.0),
                "b": (0.0, 0.0, 2.0),
                "_is_na_b[T.True]": (0.0, 1.0, 0.0),
                "c[a]": (1.0, 0.0, 1.0),
                "c[b]": (0.0, 1.0, 0.0),
            }
        )
        r = balance_util.model_matrix(s_df[["a", "b", "c"]])
        self.assertEqual(r["sample"].sort_index(axis=1), e, lazy=True)

        # Tests on a single sample with a target
        t = Sample.from_frame(
            pd.DataFrame(
                {
                    "a": (0, 1, 2, None),
                    "d": (0, 2, 2, 1),
                    "c": ("a", "b", "a", "c"),
                    "id": (1, 2, 3, 5),
                }
            )
        )

        r = balance_util.model_matrix(s, t)
        e_s = pd.DataFrame(
            {
                "a": (0.0, 1.0, 2.0),
                "_is_na_a[T.True]": (0.0, 0.0, 0.0),
                "c[a]": (1.0, 0.0, 1.0),
                "c[b]": (0.0, 1.0, 0.0),
                "c[c]": (0.0, 0.0, 0.0),
            }
        )
        e_t = pd.DataFrame(
            {
                "a": (0.0, 1.0, 2.0, 0.0),
                "_is_na_a[T.True]": (0.0, 0.0, 0.0, 1.0),
                "c[a]": (1.0, 0.0, 1.0, 0.0),
                "c[b]": (0.0, 1.0, 0.0, 0.0),
                "c[c]": (0.0, 0.0, 0.0, 1.0),
            }
        )
        self.assertEqual(r["sample"].sort_index(axis=1), e_s, lazy=True)
        self.assertEqual(r["target"].sort_index(axis=1), e_t, lazy=True)

        # Test passing DataFrames rather than Samples
        r = balance_util.model_matrix(
            pd.DataFrame({"a": (0, 1, 2), "b": (0, None, 2), "c": ("a", "b", "a")}),
            pd.DataFrame(
                {"a": (0, 1, 2, None), "d": (0, 2, 2, 1), "c": ("a", "b", "a", "c")}
            ),
        )
        self.assertEqual(r["sample"].sort_index(axis=1), e_s, lazy=True)
        self.assertEqual(r["target"].sort_index(axis=1), e_t, lazy=True)

        # Check warnings for variables not present in both
        self.assertWarnsRegexp(
            "Ignoring variables not present in all Samples",
            balance_util.model_matrix,
            s,
            t,
        )

        # Test zero rows warning:
        self.assertRaisesRegex(
            AssertionError,
            "sample must have more than zero rows",
            balance_util.model_matrix,
            pd.DataFrame(),
        )

        # Tests on a single DataFrame with bad column names
        s_df_bad_col_names = pd.DataFrame(
            {
                "a * / |a": (0, 1, 2),
                "b  b": (0, None, 2),
                "c._$c": ("a", "b", "a"),
                "id": (1, 2, 3),
            }
        )
        r = balance_util.model_matrix(s_df_bad_col_names)
        exp = ["_is_na_b__b[T.True]", "a______a", "b__b", "c___c[a]", "c___c[b]", "id"]
        self.assertEqual(r["model_matrix_columns_names"], exp)
        exp = {
            "_is_na_b__b[T.True]": {0: 0.0, 1: 1.0, 2: 0.0},
            "a______a": {0: 0.0, 1: 1.0, 2: 2.0},
            "b__b": {0: 0.0, 1: 0.0, 2: 2.0},
            "c___c[a]": {0: 1.0, 1: 0.0, 2: 1.0},
            "c___c[b]": {0: 0.0, 1: 1.0, 2: 0.0},
            "id": {0: 1.0, 1: 2.0, 2: 3.0},
        }
        self.assertEqual(r["sample"].to_dict(), exp)

        # Tests that we can handle multiple columns what would be turned to have the same column name
        s_df_bad_col_names = pd.DataFrame(
            {
                "b1 ": (0, 1, 2),
                "b1*": (0, None, 2000),
                "b1_": (3, 30, 300),
                "b1$": ["a", "b", "c"],
                "id": (1, 2, 3),
            }
        )
        r = balance_util.model_matrix(s_df_bad_col_names)
        exp = [
            "_is_na_b1_[T.True]",
            "b1_",
            "b1__1",
            "b1__2",
            "b1__3[a]",
            "b1__3[b]",
            "b1__3[c]",
            "id",
        ]
        self.assertEqual(r["model_matrix_columns_names"], exp)
        # r["sample"].to_dict()
        exp = {
            "_is_na_b1_[T.True]": {0: 0.0, 1: 1.0, 2: 0.0},
            "b1_": {0: 0.0, 1: 1.0, 2: 2.0},
            "b1__1": {0: 0.0, 1: 0.0, 2: 2000.0},
            "b1__2": {0: 3.0, 1: 30.0, 2: 300.0},
            "b1__3[a]": {0: 1.0, 1: 0.0, 2: 0.0},
            "b1__3[b]": {0: 0.0, 1: 1.0, 2: 0.0},
            "b1__3[c]": {0: 0.0, 1: 0.0, 2: 1.0},
            "id": {0: 1.0, 1: 2.0, 2: 3.0},
        }
        self.assertEqual(r["sample"].to_dict(), exp)

    def test_model_matrix_arguments(self):
        s_df = pd.DataFrame(
            {
                "a": (0, 1, 2),
                "b": (0, None, 2),
                "c": ("a", "b", "a"),
                "id": (1, 2, 3),
            }
        )
        s = Sample.from_frame(s_df)
        t = Sample.from_frame(
            pd.DataFrame(
                {
                    "a": (0, 1, 2, None),
                    "d": (0, 2, 2, 1),
                    "c": ("a", "b", "a", "c"),
                    "id": (1, 2, 3, 5),
                }
            )
        )
        # Test variables argument
        r = balance_util.model_matrix(s, variables="c")
        e = pd.DataFrame({"c[a]": (1.0, 0.0, 1.0), "c[b]": (0.0, 1.0, 0.0)})
        self.assertEqual(r["sample"], e)
        self.assertTrue(r["target"] is None)

        #  Single covariate which doesn't exist in both should raise error
        self.assertRaisesRegex(
            Exception,
            "requested variables are not in all Samples",
            balance_util.model_matrix,
            s,
            t,
            "b",
        )

        # Test add_na argument
        e = pd.DataFrame(
            {"a": (0.0, 2.0), "b": (0.0, 2.0), "c[a]": (1.0, 1.0), "c[b]": (0.0, 0.0)},
            index=(0, 2),
        )
        r = balance_util.model_matrix(s, add_na=False)
        self.assertWarnsRegexp(
            "Dropping all rows with NAs", balance_util.model_matrix, s, add_na=False
        )
        self.assertEqual(r["sample"].sort_index(axis=1), e)
        self.assertTrue(r["target"] is None)

        #  Test return_type argument
        r = balance_util.model_matrix(s, t, return_type="one")["model_matrix"]
        e_s = pd.DataFrame(
            {
                "a": (0.0, 1.0, 2.0),
                "_is_na_a[T.True]": (0.0, 0.0, 0.0),
                "c[a]": (1.0, 0.0, 1.0),
                "c[b]": (0.0, 1.0, 0.0),
                "c[c]": (0.0, 0.0, 0.0),
            }
        )
        e_t = pd.DataFrame(
            {
                "a": (0.0, 1.0, 2.0, 0.0),
                "_is_na_a[T.True]": (0.0, 0.0, 0.0, 1.0),
                "c[a]": (1.0, 0.0, 1.0, 0.0),
                "c[b]": (0.0, 1.0, 0.0, 0.0),
                "c[c]": (0.0, 0.0, 0.0, 1.0),
            }
        )
        self.assertEqual(r.sort_index(axis=1), pd.concat((e_s, e_t)), lazy=True)

        # Test return_var_type argument
        r = balance_util.model_matrix(
            s, t, return_type="one", return_var_type="dataframe"
        )["model_matrix"]
        self.assertEqual(r.sort_index(axis=1), pd.concat((e_s, e_t)), lazy=True)
        r = balance_util.model_matrix(s, t, return_type="one", return_var_type="matrix")
        self.assertEqual(
            r["model_matrix"],
            pd.concat((e_s, e_t))
            .reindex(columns=r["model_matrix_columns_names"])
            .values,
        )
        r = balance_util.model_matrix(s, t, return_type="one", return_var_type="sparse")
        self.assertEqual(
            r["model_matrix"].toarray(),
            pd.concat((e_s, e_t))
            .reindex(columns=r["model_matrix_columns_names"])
            .values,
        )
        self.assertEqual(
            r["model_matrix_columns_names"],
            ["_is_na_a[T.True]", "a", "c[a]", "c[b]", "c[c]"],
        )
        self.assertTrue(type(r["model_matrix"]) is csc_matrix)

        # Test formula argument
        self.assertEqual(
            balance_util.model_matrix(s, formula="a + b")["sample"].sort_index(axis=1),
            pd.DataFrame({"a": (0.0, 1.0, 2.0), "b": (0.0, 0.0, 2.0)}),
        )

        self.assertEqual(
            balance_util.model_matrix(s, formula="b ")["sample"].sort_index(axis=1),
            pd.DataFrame({"b": (0.0, 0.0, 2.0)}),
        )
        self.assertEqual(
            balance_util.model_matrix(s, formula="a * c ")["sample"].sort_index(axis=1),
            pd.DataFrame(
                {
                    "a": (0.0, 1.0, 2.0),
                    "a:c[T.b]": (0.0, 1.0, 0.0),
                    "c[a]": (1.0, 0.0, 1.0),
                    "c[b]": (0.0, 1.0, 0.0),
                }
            ),
        )
        self.assertEqual(
            balance_util.model_matrix(s, formula=["a", "b"])["sample"].sort_index(
                axis=1
            ),
            pd.DataFrame({"a": (0.0, 1.0, 2.0), "b": (0.0, 0.0, 2.0)}),
        )

        # Test penalty_factor argument
        self.assertEqual(
            balance_util.model_matrix(s, formula=["a", "b"])["penalty_factor"],
            np.array([1, 1]),
        )
        self.assertEqual(
            balance_util.model_matrix(s, formula=["a", "b"], penalty_factor=[1, 2])[
                "penalty_factor"
            ],
            np.array([1, 2]),
        )
        self.assertEqual(
            balance_util.model_matrix(s, formula="a+b", penalty_factor=[2])[
                "penalty_factor"
            ],
            np.array([2, 2]),
        )
        self.assertRaisesRegex(
            AssertionError,
            "penalty factor and formula must have the same length",
            balance_util.model_matrix,
            s,
            formula="a+b",
            penalty_factor=[1, 2],
        )

        # Test one_hot_encoding argument
        e = pd.DataFrame(
            {
                "C(_is_na_b, one_hot_encoding_greater_2)[True]": (0.0, 1.0, 0.0),
                "C(c, one_hot_encoding_greater_2)[b]": (0.0, 1.0, 0.0),
                "a": (0.0, 1.0, 2.0),
                "b": (0.0, 0.0, 2.0),
            }
        )
        r = balance_util.model_matrix(s, one_hot_encoding=True)
        self.assertEqual(r["sample"].sort_index(axis=1), e, lazy=True)

    def test_qcut(self):
        d = pd.Series([0, 1, 2, 3, 4])
        self.assertEqual(
            balance_util.qcut(d, 4).astype(str),
            pd.Series(
                [
                    "(-0.001, 1.0]",
                    "(-0.001, 1.0]",
                    "(1.0, 2.0]",
                    "(2.0, 3.0]",
                    "(3.0, 4.0]",
                ]
            ),
        )
        self.assertEqual(balance_util.qcut(d, 6), d)
        self.assertWarnsRegexp(
            "Not quantizing, too few values",
            balance_util.qcut,
            d,
            6,
        )

    def test_quantize(self):
        d = pd.DataFrame(np.random.rand(1000, 2))
        d = d.rename(columns={i: "ab"[i] for i in range(0, 2)})
        d["c"] = ["x"] * 1000

        r = balance_util.quantize(d, variables=("a"))
        self.assertTrue(isinstance(r["a"][0], pd.Interval))
        self.assertTrue(isinstance(r["b"][0], float))
        self.assertTrue(r["c"][0] == "x")

        r = balance_util.quantize(d)
        self.assertTrue(isinstance(r["a"][0], pd.Interval))
        self.assertTrue(isinstance(r["b"][0], pd.Interval))
        self.assertTrue(r["c"][0] == "x")

        # Test that it does not affect categorical columns
        d["d"] = pd.Categorical(["y"] * 1000)
        r = balance_util.quantize(d)
        self.assertTrue(r["d"][0] == "y")

        # Test on Series input
        r = balance_util.quantize(pd.Series(np.random.uniform(0, 1, 100)), 7)
        self.assertTrue(len(set(r.values)) == 7)

        # Test on numpy array input
        r = balance_util.quantize(np.random.uniform(0, 1, 100), 7)
        self.assertTrue(len(set(r.values)) == 7)

        # Test on single integer input
        r = balance_util.quantize(1, 1)
        self.assertTrue(len(set(r.values)) == 1)

    def test_row_pairwise_diffs(self):
        d = pd.DataFrame({"a": (1, 2, 3), "b": (-42, 8, 2)})
        e = pd.DataFrame(
            {"a": (1, 2, 3, 1, 2, 1), "b": (-42, 8, 2, 50, 44, -6)},
            index=(0, 1, 2, "1 - 0", "2 - 0", "2 - 1"),
        )
        self.assertEqual(balance_util.row_pairwise_diffs(d), e)

    def test_isarraylike(self):
        self.assertFalse(balance_util._is_arraylike(""))
        self.assertFalse(balance_util._is_arraylike("test"))
        self.assertTrue(balance_util._is_arraylike(()))
        self.assertTrue(balance_util._is_arraylike([]))
        self.assertTrue(balance_util._is_arraylike([1, 2]))
        self.assertTrue(balance_util._is_arraylike(range(10)))
        self.assertTrue(balance_util._is_arraylike(np.array([1, 2, "a"])))
        self.assertTrue(balance_util._is_arraylike(pd.Series((1, 2, 3))))

    def test_rm_mutual_nas(self):
        from balance.util import rm_mutual_nas

        self.assertEqual(rm_mutual_nas([1, 2, 3], [2, 3, None]), [[1, 2], [2.0, 3.0]])

        d = np.array((0, 1, 2))
        numpy.testing.assert_array_equal(rm_mutual_nas(d), d)

        d2 = np.array((5, 6, 7))
        numpy.testing.assert_array_equal(rm_mutual_nas(d, d2), (d, d2))

        r = rm_mutual_nas(d, d2, None)
        for i, j in zip(r, (d, d2, None)):
            numpy.testing.assert_array_equal(i, j)

        # test exceptions
        d3a = np.array((5, 6, 7, 8))
        self.assertRaisesRegex(
            ValueError,
            "All arrays must be of same length",
            rm_mutual_nas,
            d,
            d3a,
        )
        d3b = "a"
        self.assertRaisesRegex(
            ValueError,
            "All arguments must be arraylike",
            rm_mutual_nas,
            d3b,
            d3b,
        )

        d4 = np.array((np.nan, 9, -np.inf))
        e = [np.array((1,)), np.array((6,)), np.array((9,))]

        r = rm_mutual_nas(d, d2, d4)
        for i, j in zip(r, e):
            print("A", i, j)
            numpy.testing.assert_array_equal(i, j)

        r = rm_mutual_nas(d, d2, d4, None)
        for i, j in zip(r, e):
            numpy.testing.assert_array_equal(i, j)

        d5 = np.array(("a", "b", "c"))
        numpy.testing.assert_array_equal(rm_mutual_nas(d5), d5)

        e = [np.array((9,)), np.array(("b",))]
        r = rm_mutual_nas(d4, d5)
        for i, j in zip(r, e):
            numpy.testing.assert_array_equal(i, j)

        # Single arraylike
        d = np.array((0, 1, 2, None))
        numpy.testing.assert_array_equal(rm_mutual_nas(d), (0, 1, 2))
        d = np.array((0, 1, 2, np.nan))
        numpy.testing.assert_array_equal(rm_mutual_nas(d), (0, 1, 2))
        d = np.array(("a", "b", None))
        numpy.testing.assert_array_equal(rm_mutual_nas(d), ("a", "b"))
        d = np.array(("a", 1, None))
        # NOTE: In the next test we must define that `dtype=object`
        # since this dtype is preserved from d. Otherwise, using np.array(("a", 1)) will have
        # a dtype of '<U1'
        numpy.testing.assert_array_equal(
            rm_mutual_nas(d), np.array(("a", 1), dtype=object)
        )
        self.assertTrue(isinstance(rm_mutual_nas(d), np.ndarray))

        d = (0, 1, 2, None)
        numpy.testing.assert_array_equal(rm_mutual_nas(d), (0, 1, 2))
        d = ("a", "b", None)
        numpy.testing.assert_array_equal(rm_mutual_nas(d), ("a", "b"))
        d = ("a", 1, None)
        numpy.testing.assert_array_equal(rm_mutual_nas(d), ("a", 1))
        self.assertTrue(isinstance(rm_mutual_nas(d), tuple))

        # Should only accept array like or none arguments
        self.assertRaises(ValueError, rm_mutual_nas, d, "a")

        # Make sure we can deal with various np and pd arrays
        x1 = pd.array([1, 2, None, np.nan, pd.NA, 3])
        x2 = pd.array([1.1, 2, 3, None, np.nan, pd.NA])
        x3 = pd.array([1.1, 2, 3, 4, 5, 6])
        x4 = pd.array(["1.1", 2, 3, None, np.nan, pd.NA])
        x5 = pd.array(["1.1", "2", "3", None, np.nan, pd.NA], dtype="string")
        x6 = np.array([1, 2, 3.3, 4, 5, 6])
        x7 = np.array([1, 2, 3.3, 4, "5", "6"])
        x8 = [1, 2, 3.3, 4, "5", "6"]

        # The values we expect to see after using rm_mutual_nas:
        self.assertEqual(
            [list(x) for x in rm_mutual_nas(x1, x2, x3, x4, x5, x6, x7, x8)],
            [
                [1, 2],
                [1.1, 2],
                [1.1, 2.0],
                ["1.1", 2],
                ["1.1", "2"],
                [1.0, 2.0],
                ["1", "2"],
                [1, 2],
            ],
        )
        # The types before and after the na removal will remain the same:
        self.assertEqual(
            [type(x) for x in rm_mutual_nas(x1, x2, x3, x4, x5, x6, x7, x8)],
            [type(x) for x in (x1, x2, x3, x4, x5, x6, x7, x8)],
        )
        self.assertEqual(
            [type(x) for x in rm_mutual_nas(x1, x4, x5, x6, x7, x8)],
            [
                pd.core.arrays.integer.IntegerArray,
                pd.core.arrays.numpy_.PandasArray,
                pd.core.arrays.string_.StringArray,
                np.ndarray,
                np.ndarray,
                list,
            ],
        )

        # NOTE: pd.FloatingArray were only added in pandas version 1.2.0.
        # Before that, they were called PandasArray. For details, see:
        # https://pandas.pydata.org/docs/dev/reference/api/pandas.arrays.FloatingArray.html
        if pd.__version__ < "1.2.0":
            e = [
                pd.core.arrays.numpy_.PandasArray,
                pd.core.arrays.numpy_.PandasArray,
            ]
        else:
            e = [
                pd.core.arrays.floating.FloatingArray,
                pd.core.arrays.floating.FloatingArray,
            ]
        self.assertEqual([type(x) for x in rm_mutual_nas(x2, x3)], e)

        # The dtype before and after the na removal will remain the same: (only relevant for np and pd arrays)
        self.assertEqual(
            [x.dtype for x in rm_mutual_nas(x1, x2, x3, x4, x5, x6, x7)],
            [x.dtype for x in (x1, x2, x3, x4, x5, x6, x7)],
        )
        # The dtypes:
        # [Int64Dtype(),
        #  PandasDtype('object'),
        #  PandasDtype('float64'),
        #  PandasDtype('object'),
        #  StringDtype,
        #  dtype('float64'),
        #  dtype('<U32')]

        # Preserve index in pd.Series input
        x1 = pd.Series([1, 2, 3, 4])
        x2 = pd.Series([np.nan, 2, 3, 4])
        x3 = np.array([1, 2, 3, 4])
        # When there is nothing to remove, the original pd.Series will be returned with proper index:
        self.assertEqual(rm_mutual_nas(x1, x3)[0].to_dict(), {0: 1, 1: 2, 2: 3, 3: 4})
        self.assertEqual(
            rm_mutual_nas(x1.sort_values(ascending=False), x3)[0].to_dict(),
            {3: 4, 2: 3, 1: 2, 0: 1},
        )
        # Index is not changed also when na values are omitted:
        self.assertEqual(rm_mutual_nas(x1, x2)[0].to_dict(), {1: 2, 2: 3, 3: 4})
        # The order of the Series can be is changed, but the indexes remain the same
        self.assertEqual(
            rm_mutual_nas(x1.sort_values(ascending=False), x2)[0].to_dict(),
            {3: 4, 2: 3, 1: 2},
        )

    def test_choose_variables(self):
        from balance.util import choose_variables

        # For one dataframe
        self.assertEqual(
            choose_variables(pd.DataFrame({"a": [1], "b": [2]})),
            ["a", "b"],
        )
        # For two dataframes
        # Not providing variables
        self.assertEqual(
            choose_variables(
                pd.DataFrame({"a": [1], "b": [2]}),
                pd.DataFrame({"c": [1], "b": [2]}),
                variables="",
            ),
            ["b"],
        )
        self.assertEqual(
            choose_variables(
                pd.DataFrame({"a": [1], "b": [2]}), pd.DataFrame({"c": [1], "b": [2]})
            ),
            ["b"],
        )

        self.assertWarnsRegexp(
            "Ignoring variables not present in all Samples",
            choose_variables,
            pd.DataFrame({"a": [1], "b": [2]}),
            pd.DataFrame({"c": [1], "d": [2]}),
        )
        self.assertWarnsRegexp(
            "Sample and target have no variables in common",
            choose_variables,
            pd.DataFrame({"a": [1], "b": [2]}),
            pd.DataFrame({"c": [1], "d": [2]}),
        )

        self.assertEqual(
            choose_variables(
                pd.DataFrame({"a": [1], "b": [2]}), pd.DataFrame({"c": [1], "d": [2]})
            ),
            [],
        )

        with self.assertRaisesRegex(
            ValueError, "requested variables are not in all Samples"
        ):
            choose_variables(
                pd.DataFrame({"a": [1], "b": [2]}),
                pd.DataFrame({"c": [1], "b": [2]}),
                variables="a",
            )

        #  Three dataframes
        self.assertEqual(
            choose_variables(
                pd.DataFrame({"a": [1], "b": [2], "c": [2]}),
                pd.DataFrame({"c": [1], "b": [2]}),
                pd.DataFrame({"a": [1], "b": [2]}),
            ),
            ["b"],
        )

        df1 = pd.DataFrame({"a": [1], "b": [2], "c": [2]})
        df2 = pd.DataFrame({"c": [1], "b": [2]})

        with self.assertRaisesRegex(
            ValueError, "requested variables are not in all Samples: {'a'}"
        ):
            choose_variables(df1, df2, variables=("a", "b", "c"))

        # Control order
        df1 = pd.DataFrame(
            {"A": [1, 2], "B": [3, 4], "C": [5, 6], "E": [1, 1], "F": [1, 1]}
        )
        df2 = pd.DataFrame(
            {"C": [7, 8], "J": [9, 10], "B": [11, 12], "K": [1, 1], "A": [1, 1]}
        )

        self.assertEqual(
            choose_variables(df1, df2),
            ["A", "B", "C"],
        )
        self.assertEqual(
            choose_variables(df1, df2, df_for_var_order=1),
            ["C", "B", "A"],
        )
        self.assertEqual(
            choose_variables(df1, df2, variables=["B", "A"]),
            ["B", "A"],
        )

    def test_auto_spread(self):
        data = pd.DataFrame(
            {
                "id": (1, 1, 2, 2, 3),
                "key": ("a", "b", "b", "a", "a"),
                "value": (1, 1, 2, 2, 4),
            }
        )

        expected = pd.DataFrame(
            {
                "id": (1, 2, 3),
                "key_a_value": (1.0, 2.0, 4.0),
                "key_b_value": (1.0, 2.0, np.nan),
            },
            columns=("id", "key_a_value", "key_b_value"),
        )
        self.assertEqual(expected, balance_util.auto_spread(data))

        data = pd.DataFrame(
            {
                "id": (1, 1, 2, 2, 3),
                "key": ("a", "b", "b", "a", "a"),
                "value": (1, 1, 2, 2, 4),
                "other_value": (2, 2, 4, 4, 6),
            }
        )

        self.assertEqual(
            expected, balance_util.auto_spread(data, features=["key", "value"])
        )

        expected = pd.DataFrame(
            {
                "id": (1, 2, 3),
                "key_a_value": (1.0, 2.0, 4.0),
                "key_b_value": (1.0, 2.0, np.nan),
                "key_a_other_value": (2.0, 4.0, 6.0),
                "key_b_other_value": (2.0, 4.0, np.nan),
            },
            columns=(
                "id",
                "key_a_other_value",
                "key_b_other_value",
                "key_a_value",
                "key_b_value",
            ),
        )
        self.assertEqual(expected, balance_util.auto_spread(data), lazy=True)

        data = pd.DataFrame(
            {
                "id": (1, 1, 2, 2, 3),
                "key": ("a", "a", "c", "d", "a"),
                "value": (1, 1, 2, 4, 1),
            }
        )
        self.assertWarnsRegexp("no unique groupings", balance_util.auto_spread, data)

    def test_auto_spread_multiple_groupings(self):
        # Multiple possible groupings
        data = pd.DataFrame(
            {
                "id": (1, 1, 2, 2, 3),
                "key": ("a", "b", "b", "a", "a"),
                "value": (1, 3, 2, 4, 1),
            }
        )
        expected = pd.DataFrame(
            {
                "id": (1, 2, 3),
                "key_a_value": (1.0, 4.0, 1.0),
                "key_b_value": (3.0, 2.0, np.nan),
            },
            columns=("id", "key_a_value", "key_b_value"),
        )
        self.assertEqual(expected, balance_util.auto_spread(data))
        self.assertWarnsRegexp("2 possible groupings", balance_util.auto_spread, data)

    def test_auto_aggregate(self):
        r = balance_util.auto_aggregate(
            pd.DataFrame(
                {"x": [1, 2, 3, 4], "y": [1, 1, 1, np.nan], "id": [1, 1, 2, 3]}
            )
        )
        e = pd.DataFrame({"id": [1, 2, 3], "x": [3, 3, 4], "y": [2, 1, np.nan]})

        self.assertEqual(r, e, lazy=True)

        self.assertRaises(
            ValueError,
            balance_util.auto_aggregate,
            pd.DataFrame({"b": ["a", "b", "b"], "id": [1, 1, 2]}),
        )

        self.assertRaises(
            ValueError,
            balance_util.auto_aggregate,
            r,
            None,
            "id2",
        )

        self.assertRaises(
            ValueError,
            balance_util.auto_aggregate,
            r,
            None,
            aggfunc="not_sum",
        )

    def test_fct_lump(self):
        # Count above the threshold, value preserved
        s = pd.Series(["a"] * 95 + ["b"] * 5)
        self.assertEqual(balance_util.fct_lump(s), s)

        # Move the threshold up
        self.assertEqual(
            balance_util.fct_lump(s, 0.10),
            pd.Series(["a"] * 95 + ["_lumped_other"] * 5),
        )

        # Default threshold, slightly below number of values
        self.assertEqual(
            balance_util.fct_lump(pd.Series(["a"] * 96 + ["b"] * 4)),
            pd.Series(["a"] * 96 + ["_lumped_other"] * 4),
        )

        # Multiple categories combined
        self.assertEqual(
            balance_util.fct_lump(pd.Series(["a"] * 96 + ["b"] * 2 + ["c"] * 2)),
            pd.Series(["a"] * 96 + ["_lumped_other"] * 4),
        )

        # Category already called '_lumped_other' is handled
        self.assertEqual(
            balance_util.fct_lump(pd.Series(["a"] * 96 + ["_lumped_other"] * 4)),
            pd.Series(["a"] * 96 + ["_lumped_other_lumped_other"] * 4),
        )

        # Categorical series type
        self.assertEqual(
            balance_util.fct_lump(pd.Series(["a"] * 96 + ["b"] * 4, dtype="category")),
            pd.Series(["a"] * 96 + ["_lumped_other"] * 4),
        )

        # Categorical model matrix is equal to string model matrix
        # Load and process test data
        from sklearn import datasets

        wine_df = pd.DataFrame(datasets.load_wine().data)
        wine_df.columns = datasets.load_wine().feature_names
        wine_df = wine_df.rename(
            columns={"od280/od315_of_diluted_wines": "od280_od315_of_diluted_wines"}
        )
        wine_df["id"] = pd.Series(
            range(1, len(wine_df) + 1)
        )  # Create a fake id variable, required by balance
        wine_df.alcohol = pd.cut(
            wine_df.alcohol, bins=[0, 11, 11.5, 12, 12.5, 13, 13.5, 14, 14.5, 100]
        )

        # Create a version of the dataset that treats factor variable as character
        wine_df_copy = wine_df.copy(deep=True)
        wine_df_copy.alcohol = wine_df_copy.alcohol.astype("object")

        # Split into "survey" and "population" datasets based on "wine class"
        wine_class = pd.Series(datasets.load_wine().target)
        wine_survey = Sample.from_frame(wine_df.loc[wine_class == 0, :])
        wine_pop = Sample.from_frame(wine_df.loc[wine_class != 0, :])
        wine_survey = wine_survey.set_target(wine_pop)
        wine_survey_copy = Sample.from_frame(wine_df_copy.loc[wine_class == 0, :])
        wine_pop_copy = Sample.from_frame(wine_df_copy.loc[wine_class != 0, :])
        wine_survey_copy = wine_survey_copy.set_target(wine_pop_copy)

        # Generate weights
        output_cat_var = wine_survey.adjust(
            transformations={
                "alcohol": lambda x: balance_util.fct_lump(x, prop=0.05),
                "flavanoids": balance_util.quantize,
                "total_phenols": balance_util.quantize,
                "nonflavanoid_phenols": balance_util.quantize,
                "color_intensity": balance_util.quantize,
                "hue": balance_util.quantize,
                "ash": balance_util.quantize,
                "alcalinity_of_ash": balance_util.quantize,
                "malic_acid": balance_util.quantize,
                "magnesium": balance_util.quantize,
            }
        )

        output_string_var = wine_survey_copy.adjust(
            transformations={
                "alcohol": lambda x: balance_util.fct_lump(x, prop=0.05),
                "flavanoids": balance_util.quantize,
                "total_phenols": balance_util.quantize,
                "nonflavanoid_phenols": balance_util.quantize,
                "color_intensity": balance_util.quantize,
                "hue": balance_util.quantize,
                "ash": balance_util.quantize,
                "alcalinity_of_ash": balance_util.quantize,
                "malic_acid": balance_util.quantize,
                "magnesium": balance_util.quantize,
            }
        )

        # Check that model coefficients are identical in categorical and string variable coding
        self.assertEqual(
            output_cat_var.model()["perf"]["coefs"],
            output_string_var.model()["perf"]["coefs"],
        )

    def test_fct_lump_by(self):
        # test by argument works
        s = pd.Series([1, 1, 1, 2, 3, 1, 2])
        by = pd.Series(["a", "a", "a", "a", "a", "b", "b"])
        self.assertEqual(
            balance_util.fct_lump_by(s, by, 0.5),
            pd.Series([1, 1, 1, "_lumped_other", "_lumped_other", 1, 2]),
        )

        # test case where all values in 'by' are the same
        s = pd.Series([1, 1, 1, 2, 3, 1, 2])
        by = pd.Series(["a", "a", "a", "a", "a", "a", "a"])
        self.assertEqual(
            balance_util.fct_lump_by(s, by, 0.5),
            pd.Series(
                [1, 1, 1, "_lumped_other", "_lumped_other", 1, "_lumped_other"],
                name="a",
            ),
        )

        # test fct_lump_by doesn't affect indices when combining dataframes
        s = pd.DataFrame({"d": [1, 1, 1], "e": ["a1", "a2", "a1"]}, index=(0, 6, 7))
        t = pd.DataFrame(
            {"d": [2, 3, 1, 2], "e": ["a2", "a2", "a1", "a2"]}, index=(0, 1, 2, 3)
        )
        df = pd.concat([s, t])
        r = balance_util.fct_lump_by(df.d, df.e, 0.5)
        e = pd.Series(
            [1, "_lumped_other", 1, 2, "_lumped_other", 1, 2],
            index=(0, 6, 7, 0, 1, 2, 3),
            name="d",
        )
        self.assertEqual(r, e)

    def test_one_hot_encoding_greater_2(self):
        from balance.util import one_hot_encoding_greater_2  # noqa

        d = {
            "a": ["a1", "a2", "a1", "a1"],
            "b": ["b1", "b2", "b3", "b3"],
            "c": ["c1", "c1", "c1", "c1"],
        }
        df = pd.DataFrame(data=d)

        res = dmatrix("C(a, one_hot_encoding_greater_2)", df, return_type="dataframe")
        expected = {
            "Intercept": [1.0, 1.0, 1.0, 1.0],
            "C(a, one_hot_encoding_greater_2)[a2]": [0.0, 1.0, 0.0, 0.0],
        }
        expected = pd.DataFrame(data=expected)
        self.assertEqual(res, expected)

        res = dmatrix("C(b, one_hot_encoding_greater_2)", df, return_type="dataframe")
        expected = {
            "Intercept": [1.0, 1.0, 1.0, 1.0],
            "C(b, one_hot_encoding_greater_2)[b1]": [1.0, 0.0, 0.0, 0.0],
            "C(b, one_hot_encoding_greater_2)[b2]": [0.0, 1.0, 0.0, 0.0],
            "C(b, one_hot_encoding_greater_2)[b3]": [0.0, 0.0, 1.0, 1.0],
        }
        expected = pd.DataFrame(data=expected)
        self.assertEqual(res, expected)

        res = dmatrix("C(c, one_hot_encoding_greater_2)", df, return_type="dataframe")
        expected = {
            "Intercept": [1.0, 1.0, 1.0, 1.0],
            "C(c, one_hot_encoding_greater_2)[c1]": [1.0, 1.0, 1.0, 1.0],
        }
        expected = pd.DataFrame(data=expected)
        self.assertEqual(res, expected)

    def test_truncate_text(self):
        self.assertEqual(
            balance_util._truncate_text("a" * 6, length=5), "a" * 5 + "..."
        )
        self.assertEqual(balance_util._truncate_text("a" * 4, length=5), "a" * 4)
        self.assertEqual(balance_util._truncate_text("a" * 5, length=5), "a" * 5)

    def test__dict_intersect(self):
        d1 = {"a": 1, "b": 2}
        d2 = {"c": 3, "b": 2222}
        self.assertEqual(balance_util._dict_intersect(d1, d2), {"b": 2})

    def test__astype_in_df_from_dtypes(self):
        df = pd.DataFrame({"id": ("1", "2"), "a": (1.0, 2.0), "weight": (1.0, 2.0)})
        df_orig = pd.DataFrame(
            {"id": (1, 2), "a": (1, 2), "forest": ("tree", "banana")}
        )

        self.assertEqual(
            df.dtypes.to_dict(),
            {"id": dtype("O"), "a": dtype("float64"), "weight": dtype("float64")},
        )
        self.assertEqual(
            df_orig.dtypes.to_dict(),
            {"id": dtype("int64"), "a": dtype("int64"), "forest": dtype("O")},
        )

        df_fixed = balance_util._astype_in_df_from_dtypes(df, df_orig.dtypes)
        self.assertEqual(
            df_fixed.dtypes.to_dict(),
            {"id": dtype("int64"), "a": dtype("int64"), "weight": dtype("float64")},
        )

    def test__true_false_str_to_bool(self):
        self.assertFalse(balance_util._true_false_str_to_bool("falsE"))
        self.assertTrue(balance_util._true_false_str_to_bool("TrUe"))
        with self.assertRaisesRegex(
            ValueError,
            "Banana is not an accepted value, please pass either 'True' or 'False'*",
        ):
            balance_util._true_false_str_to_bool("Banana")

    def test__are_dtypes_equal(self):
        df1 = pd.DataFrame({"int": np.arange(5), "flt": np.random.randn(5)})
        df2 = pd.DataFrame({"flt": np.random.randn(5), "int": np.random.randn(5)})
        df11 = pd.DataFrame(
            {"int": np.arange(5), "flt": np.random.randn(5), "miao": np.random.randn(5)}
        )

        self.assertTrue(
            balance_util._are_dtypes_equal(df1.dtypes, df1.dtypes)["is_equal"]
        )
        self.assertFalse(
            balance_util._are_dtypes_equal(df1.dtypes, df2.dtypes)["is_equal"]
        )
        self.assertFalse(
            balance_util._are_dtypes_equal(df11.dtypes, df2.dtypes)["is_equal"]
        )

    def test__warn_of_df_dtypes_change(self):
        df = pd.DataFrame({"int": np.arange(5), "flt": np.random.randn(5)})
        new_df = deepcopy(df)
        new_df.int = new_df.int.astype(float)
        new_df.flt = new_df.flt.astype(int)

        self.assertWarnsRegexp(
            "The dtypes of new_df were changed from the original dtypes of the input df, here are the differences - ",
            balance_util._warn_of_df_dtypes_change,
            df.dtypes,
            new_df.dtypes,
        )

    def test__make_df_column_names_unique(self):
        # Sample DataFrame with duplicate column names
        data = {
            "A": [1, 2, 3],
            "B": [4, 5, 6],
            "A2": [7, 8, 9],
            "C": [10, 11, 12],
        }

        df1 = pd.DataFrame(data)
        df1.columns = ["A", "B", "A", "A"]

        # TODO: understand in the future why the names here appear to be consistent while when using the function in
        # `model_matrix` it does not appear to work.
        self.assertEqual(
            balance_util._make_df_column_names_unique(df1).to_dict(),
            {
                "A": {0: 1, 1: 2, 2: 3},
                "B": {0: 4, 1: 5, 2: 6},
                "A_1": {0: 7, 1: 8, 2: 9},
                "A_2": {0: 10, 1: 11, 2: 12},
            },
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import balance.testutil

import numpy as np
import pandas as pd
from balance.sample_class import Sample
from balance.weighting_methods.poststratify import poststratify


class Testpoststratify(
    balance.testutil.BalanceTestCase,
):
    def test_poststratify(self):
        s = pd.DataFrame(
            {
                "a": (0, 1, 0, 1),
                "c": ("a", "a", "b", "b"),
            },
        )
        s_weights = pd.Series([4, 2, 1, 1])
        t = s
        t_weights = pd.Series([4, 2, 2, 8])
        result = poststratify(
            sample_df=s, sample_weights=s_weights, target_df=t, target_weights=t_weights
        )["weight"]
        self.assertEqual(result, t_weights.astype("float64"))

        # same example when dataframe of elements are all related to weights of one
        s = pd.DataFrame(
            {
                "a": (0, 0, 0, 0, 1, 1, 0, 1),
                "c": ("a", "a", "a", "a", "a", "a", "b", "b"),
            },
        )
        s_weights = pd.Series([1, 1, 1, 1, 1, 1, 1, 1])
        result = poststratify(
            sample_df=s, sample_weights=s_weights, target_df=t, target_weights=t_weights
        )["weight"]
        self.assertEqual(result, pd.Series((1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 2.0, 8.0)))

        # same example with normalized weights
        s = pd.DataFrame(
            {
                "a": (0, 1, 0, 1),
                "c": ("a", "a", "b", "b"),
            },
        )
        s_weights = pd.Series([1 / 2, 1 / 4, 1 / 8, 1 / 8])
        result = poststratify(
            sample_df=s, sample_weights=s_weights, target_df=t, target_weights=t_weights
        )["weight"]
        self.assertEqual(result, t_weights.astype("float64"))

        # test through adjustment
        # TODO: test the previous example through adjustment as well
        sample = Sample.from_frame(
            df=pd.DataFrame(
                {
                    "a": (1, 2, 3, 1),
                    "b": (-42, 8, 2, -42),
                    "o": (7, 8, 9, 10),
                    "c": ("x", "y", "z", "x"),
                    "id": (1, 2, 3, 4),
                    "w": (0.5, 2, 1, 1),
                }
            ),
            id_column="id",
            weight_column="w",
            outcome_columns="o",
        )
        target = Sample.from_frame(
            pd.DataFrame(
                {
                    "a": (1, 2, 3),
                    "b": (-42, 8, 2),
                    "c": ("x", "y", "z"),
                    "id": (1, 2, 3),
                    "w": (2, 0.5, 1),
                }
            ),
            id_column="id",
            weight_column="w",
        )
        result = sample.adjust(target, method="poststratify", transformations=None)
        expected = pd.Series(
            (
                (2 / 1.5 * 0.5),
                (0.5 / 2 * 2),
                (1 / 1 * 1),
                (2 / 1.5 * 1),
            )
        )
        self.assertEqual(expected, result.weights().df.iloc[:, 0].values)

    def test_poststratify_variables_arg(self):
        s = pd.DataFrame(
            {
                "a": (0, 1, 0, 1),
                "c": ("a", "a", "b", "b"),
            },
        )
        s_weights = pd.Series([4, 2, 2, 3])
        t = s
        t_weights = pd.Series([4, 2, 2, 8])
        result = poststratify(
            sample_df=s,
            sample_weights=s_weights,
            target_df=t,
            target_weights=t_weights,
            variables=["a"],
        )["weight"]
        self.assertEqual(result, pd.Series([4.0, 4.0, 2.0, 6.0]))

    def test_poststratify_transformations(self):
        # for numeric
        size = 10000
        s = pd.DataFrame({"age": np.random.uniform(0, 1, size)})
        tmp = int(size * 0.2)
        t = pd.DataFrame(
            {
                "age": np.concatenate(
                    (
                        np.random.uniform(0, 0.4, tmp),
                        np.random.uniform(0.4, 1, size - tmp),
                    )
                )
            }
        )
        result = poststratify(
            sample_df=s,
            sample_weights=pd.Series([1] * size),
            target_df=t,
            target_weights=pd.Series([1] * size),
        )["weight"]

        # age>0.4 has 4 times as many people than age <0.4 in the target
        # Check that the weights come out as 0.2 and 0.8
        eps = 0.05
        self.assertTrue(abs(result[s.age < 0.4].sum() / size - 0.2) < eps)
        self.assertTrue(abs(result[s.age >= 0.4].sum() / size - 0.8) < eps)

        # for strings
        size = 10000
        s = pd.DataFrame(
            {"x": np.random.choice(("a", "b", "c"), size, p=(0.95, 0.035, 0.015))}
        )
        t = pd.DataFrame(
            {"x": np.random.choice(("a", "b", "c"), size, p=(0.95, 0.015, 0.035))}
        )
        result = poststratify(
            sample_df=s,
            sample_weights=pd.Series([1] * size),
            target_df=t,
            target_weights=pd.Series([1] * size),
        )["weight"]

        # Output weights should ignore the difference between values 'b' and 'c'
        # since these are combined in default transformations (into '_lumped_other').
        # Hence their frequency would be as in sample
        eps = 0.05
        self.assertTrue(abs(result[s.x == "a"].sum() / size - 0.95) < eps)
        self.assertTrue(abs(result[s.x == "b"].sum() / size - 0.035) < eps)
        self.assertTrue(abs(result[s.x == "c"].sum() / size - 0.015) < eps)

    def test_poststratify_exceptions(self):
        # column with name weight
        s = pd.DataFrame(
            {
                "weight": (0, 1, 0, 1),
                "c": ("a", "a", "b", "b"),
            },
        )
        s_weights = pd.Series([4, 2, 1, 1])
        t = pd.DataFrame(
            {
                "a": (0, 1, 0, 1),
                "c": ("a", "a", "b", "b"),
            },
        )
        t_weights = pd.Series([4, 2, 2, 8])
        with self.assertRaisesRegex(
            ValueError,
            "weight can't be a name of a column in sample or target when applying poststratify",
        ):
            poststratify(s, s_weights, t, t_weights)
        with self.assertRaisesRegex(
            ValueError,
            "weight can't be a name of a column in sample or target when applying poststratify",
        ):
            poststratify(t, t_weights, s, s_weights)

        # not all sample cells are in target
        s = pd.DataFrame(
            {
                "a": ("x", "y"),
                "b": (0, 1),
            },
        )
        s_weights = pd.Series([1] * 2)
        t = pd.DataFrame(
            {
                "a": ("x", "x", "y"),
                "b": (0, 1, 0),
            },
        )
        t_weights = pd.Series([2] * 3)
        with self.assertRaisesRegex(
            ValueError, "all combinations of cells in sample_df must be in target_df"
        ):
            poststratify(s, s_weights, t, t_weights)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

import os.path

import tempfile

import balance.testutil

import numpy as np
import pandas as pd
from balance.cli import _float_or_none, BalanceCLI, make_parser
from numpy import dtype


# TODO: this should probably be generalized (so the code below could be made cleaner)
def check_some_flags(flag=True, the_flag_str="--skip_standardize_types"):
    with tempfile.TemporaryDirectory() as temp_dir, tempfile.NamedTemporaryFile(
        "w", suffix=".csv", delete=False
    ) as in_file:

        in_contents = (
            "x,y,is_respondent,id,weight\n"
            + ("1.0,50,1,1,1\n" * 1000)
            + ("2.0,60,0,1,1\n" * 1000)
        )
        in_file.write(in_contents)
        in_file.close()
        out_file = os.path.join(temp_dir, "out.csv")

        parser = make_parser()

        args_to_parse = [
            "--input_file",
            in_file.name,
            "--output_file",
            out_file,
            "--covariate_columns",
            "x,y",
        ]
        if flag:
            args_to_parse.append(the_flag_str)
        args = parser.parse_args(args_to_parse)
        cli = BalanceCLI(args)
        cli.update_attributes_for_main_used_by_adjust()
        cli.main()

        pd_in = pd.read_csv(in_file.name)
        pd_out = pd.read_csv(out_file)
    return {"pd_in": pd_in, "pd_out": pd_out}


class TestCli(
    balance.testutil.BalanceTestCase,
):
    def test_cli_help(self):
        #  Just make sure it doesn't fail
        try:
            parser = make_parser()
            args = parser.parse_args(["--help"])
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()
        except SystemExit as e:
            if e.code == 0:
                pass
            else:
                raise e

    def test_cli_float_or_none(self):
        self.assertEqual(_float_or_none(None), None)
        self.assertEqual(_float_or_none("None"), None)
        self.assertEqual(_float_or_none("13.37"), 13.37)

    def test_cli_succeed_on_weighting_failure(self):
        with tempfile.TemporaryDirectory() as temp_dir, tempfile.NamedTemporaryFile(
            "w", suffix=".csv", delete=False
        ) as in_file:

            in_contents = "x,y,is_respondent,id,weight\n" "a,b,1,1,1\n" "a,b,0,1,1"
            in_file.write(in_contents)
            in_file.close()
            out_file = os.path.join(temp_dir, "out.csv")

            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    in_file.name,
                    "--output_file",
                    out_file,
                    "--covariate_columns",
                    "x,y",
                ]
            )
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            self.assertRaisesRegex(Exception, "glmnet", cli.main)

            args = parser.parse_args(
                [
                    "--input_file",
                    in_file.name,
                    "--output_file",
                    out_file,
                    "--covariate_columns",
                    "x,y",
                    "--succeed_on_weighting_failure",
                ]
            )
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()

            self.assertTrue(os.path.isfile(out_file))

    def test_cli_works(self):
        with tempfile.TemporaryDirectory() as temp_dir, tempfile.NamedTemporaryFile(
            "w", suffix=".csv", delete=False
        ) as in_file:

            in_contents = (
                "x,y,is_respondent,id,weight\n"
                + ("a,b,1,1,1\n" * 1000)
                + ("c,b,0,1,1\n" * 1000)
            )
            in_file.write(in_contents)
            in_file.close()
            out_file = os.path.join(temp_dir, "out.csv")
            diagnostics_out_file = os.path.join(temp_dir, "diagnostics_out.csv")

            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    in_file.name,
                    "--output_file",
                    out_file,
                    "--diagnostics_output_file",
                    diagnostics_out_file,
                    "--covariate_columns",
                    "x,y",
                ]
            )
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()

            self.assertTrue(os.path.isfile(out_file))
            self.assertTrue(os.path.isfile(diagnostics_out_file))

    def test_cli_works_with_row_column_filters(self):
        import os

        # TODO: remove imports (already imported in the beginning of the file)
        import tempfile

        import pandas as pd

        with tempfile.TemporaryDirectory() as temp_dir, tempfile.NamedTemporaryFile(
            "w", suffix=".csv", delete=False
        ) as in_file:
            # # create temp folder / file
            # in_file = tempfile.NamedTemporaryFile(
            #             "w", suffix=".csv", delete=False
            #         )
            # temp_dir = tempfile.TemporaryDirectory().name

            in_contents = (
                "x,y,z,is_respondent,id,weight\n"
                + ("a,b,g,1,1,1\n" * 1000)
                + ("c,b,g,1,1,1\n" * 1000)
                + ("c,b,g,0,1,1\n" * 1000)
                + ("a,d,n,1,2,1\n" * 1000)
            )
            in_file.write(in_contents)
            in_file.close()
            # pd.read_csv(in_file)

            out_file = os.path.join(temp_dir, "out.csv")
            diagnostics_out_file = os.path.join(temp_dir, "diagnostics_out.csv")

            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    in_file.name,
                    "--output_file",
                    out_file,
                    "--diagnostics_output_file",
                    diagnostics_out_file,
                    "--covariate_columns",
                    "x,y,z",
                    "--rows_to_keep_for_diagnostics",
                    "z == 'g'",  # filtering condition
                    "--covariate_columns_for_diagnostics",
                    "x,y",  # ignoring z
                ]
            )

            # run cli
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()

            # get the files created from cli to pandas to check them
            pd_in_file = pd.read_csv(in_file.name)
            pd_out_file = pd.read_csv(out_file)
            pd_diagnostics_out_file = pd.read_csv(diagnostics_out_file)

            # test stuff
            # Make sure we indeed got all the output files
            self.assertTrue(os.path.isfile(out_file))
            self.assertTrue(os.path.isfile(diagnostics_out_file))

            # the original file had 4000 rows (1k target and 3k sample)
            self.assertEqual(pd_in_file.shape, (4000, 6))
            # the cli output includes only the panel (NOT the target)
            #   it also includes all the original columns
            self.assertEqual(pd_out_file.shape, (3000, 6))
            self.assertEqual(pd_out_file.is_respondent.mean(), 1)
            # the diagnostics file shows it was calculated on only 2k panelists (as required from the condition)
            ss = pd_diagnostics_out_file.eval(
                "(metric == 'size') & (var == 'sample_obs')"
            )
            self.assertEqual(int(pd_diagnostics_out_file[ss]["val"]), 2000)

            # verify we get diagnostics only for x and y, and not z
            ss = pd_diagnostics_out_file.eval("metric == 'covar_main_asmd_adjusted'")
            output = pd_diagnostics_out_file[ss]["var"].to_list()
            expected = ["x", "y", "mean(asmd)"]
            self.assertEqual(output, expected)

        # # remove temp folder / file
        # os.unlink(in_file.name)
        # import shutil
        # shutil.rmtree(temp_dir)

    def test_cli_empty_input(self):
        with tempfile.TemporaryDirectory() as temp_dir, tempfile.NamedTemporaryFile(
            "w", suffix=".csv", delete=False
        ) as in_file:

            in_contents = "x,y,is_respondent,id,weight\n"
            in_file.write(in_contents)
            in_file.close()
            out_file = os.path.join(temp_dir, "out.csv")
            diagnostics_out_file = os.path.join(temp_dir, "diagnostics_out.csv")

            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    in_file.name,
                    "--output_file",
                    out_file,
                    "--diagnostics_output_file",
                    diagnostics_out_file,
                    "--covariate_columns",
                    "x,y",
                ]
            )
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()

            self.assertTrue(os.path.isfile(out_file))
            self.assertTrue(os.path.isfile(diagnostics_out_file))

    def test_cli_empty_input_keep_row(self):
        with tempfile.TemporaryDirectory() as temp_dir, tempfile.NamedTemporaryFile(
            "w", suffix=".csv", delete=False
        ) as in_file:

            in_contents = "x,y,is_respondent,id,weight,keep_row,batch_column\n"
            in_file.write(in_contents)
            in_file.close()
            out_file = os.path.join(temp_dir, "out.csv")
            diagnostics_out_file = os.path.join(temp_dir, "diagnostics_out.csv")

            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    in_file.name,
                    "--output_file",
                    out_file,
                    "--diagnostics_output_file",
                    diagnostics_out_file,
                    "--covariate_columns",
                    "x,y",
                    "--keep_row",
                    "keep_row",
                    "--batch_columns",
                    "batch_column",
                ]
            )
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()

            self.assertTrue(os.path.isfile(out_file))
            self.assertTrue(os.path.isfile(diagnostics_out_file))

    def test_cli_sep_works(self):
        import os
        import tempfile

        import pandas as pd

        with tempfile.TemporaryDirectory() as temp_dir, tempfile.NamedTemporaryFile(
            "w", suffix=".csv", delete=False
        ) as in_file:
            # # create temp folder / file
            # in_file = tempfile.NamedTemporaryFile(
            #             "w", suffix=".csv", delete=False
            #         )
            # temp_dir = tempfile.TemporaryDirectory().name

            in_contents = (
                "x,y,z,is_respondent,id,weight\n"
                + ("a,b,g,1,1,1\n" * 1000)
                + ("c,b,g,1,1,1\n" * 1000)
                + ("c,b,g,0,1,1\n" * 1000)
                + ("a,d,n,1,2,1\n" * 1000)
            )
            in_file.write(in_contents)
            in_file.close()
            # pd.read_csv(in_file)

            out_file = os.path.join(temp_dir, "out.csv")
            diagnostics_out_file = os.path.join(temp_dir, "diagnostics_out.csv")

            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    in_file.name,
                    "--output_file",
                    out_file,
                    "--diagnostics_output_file",
                    diagnostics_out_file,
                    "--covariate_columns",
                    "x,y,z",
                    "--sep_output_file",
                    ";",
                    "--sep_diagnostics_output_file",
                    ";",
                ]
            )

            # run cli
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()

            # get the files created from cli to pandas to check them
            pd_in_file = pd.read_csv(in_file.name)
            pd_out_file = pd.read_csv(out_file, sep=";")
            pd_diagnostics_out_file = pd.read_csv(diagnostics_out_file, sep=";")

            # test stuff
            # Make sure we indeed got all the output files
            self.assertTrue(os.path.isfile(out_file))
            self.assertTrue(os.path.isfile(diagnostics_out_file))

            # the original file had 4000 rows (1k target and 3k sample)
            self.assertEqual(pd_in_file.shape, (4000, 6))
            # the cli output includes only the panel (NOT the target)
            #   it also includes all the original columns
            self.assertEqual(pd_out_file.shape, (3000, 6))
            self.assertEqual(pd_out_file.is_respondent.mean(), 1)
            # the diagnostics file shows it was calculated on all 3k panelists
            ss = pd_diagnostics_out_file.eval(
                "(metric == 'size') & (var == 'sample_obs')"
            )
            self.assertEqual(int(pd_diagnostics_out_file[ss]["val"]), 3000)

        # # remove temp folder / file
        # os.unlink(in_file.name)
        # import shutil
        # shutil.rmtree(temp_dir)

    def test_cli_sep_input_works(self):
        import os
        import tempfile

        import pandas as pd

        with tempfile.TemporaryDirectory() as temp_dir, tempfile.NamedTemporaryFile(
            "w", suffix=".tsv", delete=False
        ) as in_file:

            in_contents = (
                "x\ty\tz\tis_respondent\tid\tweight\n"
                + ("a\tb\tg\t1\t1\t1\n" * 1000)
                + ("c\tb\tg\t1\t1\t1\n" * 1000)
                + ("c\tb\tg\t0\t1\t1\n" * 1000)
                + ("a\td\tn\t1\t2\t1\n" * 1000)
            )
            in_file.write(in_contents)
            in_file.close()

            out_file = os.path.join(temp_dir, "out.csv")
            diagnostics_out_file = os.path.join(temp_dir, "diagnostics_out.csv")

            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    in_file.name,
                    "--output_file",
                    out_file,
                    "--diagnostics_output_file",
                    diagnostics_out_file,
                    "--covariate_columns",
                    "x,y,z",
                    "--sep_input_file",
                    "\t",
                ]
            )

            # run cli
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()

            # get the files created from cli to pandas to check them
            pd_in_file = pd.read_csv(in_file.name, sep="\t")
            pd_out_file = pd.read_csv(out_file)
            pd_diagnostics_out_file = pd.read_csv(diagnostics_out_file)

            # test stuff
            # Make sure we indeed got all the output files
            self.assertTrue(os.path.isfile(out_file))
            self.assertTrue(os.path.isfile(diagnostics_out_file))

            # the original file had 4000 rows (1k target and 3k sample)
            self.assertEqual(pd_in_file.shape, (4000, 6))
            # the cli output includes only the panel (NOT the target)
            #   it also includes all the original columns
            self.assertEqual(pd_out_file.shape, (3000, 6))
            self.assertEqual(pd_out_file.is_respondent.mean(), 1)
            # the diagnostics file shows it was calculated on all 3k panelists
            ss = pd_diagnostics_out_file.eval(
                "(metric == 'size') & (var == 'sample_obs')"
            )
            self.assertEqual(int(pd_diagnostics_out_file[ss]["val"]), 3000)

    def test_cli_short_arg_names_works(self):
        # Some users used only partial arg names for their pipelines
        # This is not good practice, but we'd like to not break these pipelines.
        # Hence, this test verifies new arguments would still be backward compatible
        import os
        import tempfile

        import pandas as pd

        with tempfile.TemporaryDirectory() as temp_dir, tempfile.NamedTemporaryFile(
            "w", suffix=".csv", delete=False
        ) as in_file:

            in_contents = (
                "x,y,z,is_respondent,id,weight\n"
                + ("a,b,g,1,1,1\n" * 1000)
                + ("c,b,g,1,1,1\n" * 1000)
                + ("c,b,g,0,1,1\n" * 1000)
                + ("a,d,n,1,2,1\n" * 1000)
            )
            in_file.write(in_contents)
            in_file.close()
            # pd.read_csv(in_file)

            out_file = os.path.join(temp_dir, "out.csv")
            diagnostics_out_file = os.path.join(temp_dir, "diagnostics_out.csv")

            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input",
                    in_file.name,
                    "--output",  # instead of output_file
                    out_file,
                    "--diagnostics_output_file",
                    diagnostics_out_file,
                    "--covariate_columns",
                    "x,y,z",
                    "--sep_output",
                    ";",
                    "--sep_diagnostics",
                    ";",
                ]
            )

            # run cli
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()

            # get the files created from cli to pandas to check them
            pd_in_file = pd.read_csv(in_file.name)
            pd_out_file = pd.read_csv(out_file, sep=";")
            pd_diagnostics_out_file = pd.read_csv(diagnostics_out_file, sep=";")

            # test stuff
            # Make sure we indeed got all the output files
            self.assertTrue(os.path.isfile(out_file))
            self.assertTrue(os.path.isfile(diagnostics_out_file))

            # the original file had 4000 rows (1k target and 3k sample)
            self.assertEqual(pd_in_file.shape, (4000, 6))
            # the cli output includes only the panel (NOT the target)
            #   it also includes all the original columns
            self.assertEqual(pd_out_file.shape, (3000, 6))
            self.assertEqual(pd_out_file.is_respondent.mean(), 1)
            # the diagnostics file shows it was calculated on all 3k panelists
            ss = pd_diagnostics_out_file.eval(
                "(metric == 'size') & (var == 'sample_obs')"
            )
            self.assertEqual(int(pd_diagnostics_out_file[ss]["val"]), 3000)

    # TODO: Add tests for max_de
    # TODO: Add tests for weight_trimming_mean_ratio

    def test_method_works(self):
        # TODO: ideally we'll have the example outside, and a different function for each of the methods (ipw, cbps, raking)
        np.random.seed(2021)
        n_sample = 1000
        n_target = 2000
        sample_df = pd.DataFrame(
            {
                "age": np.random.uniform(0, 100, n_sample),
                "gender": np.random.choice((1, 2, 3, 4), n_sample),
                "id": range(n_sample),
                "weight": pd.Series((1,) * n_sample),
            }
        )
        sample_df["is_respondent"] = True
        target_df = pd.DataFrame(
            {
                "age": np.random.uniform(0, 100, n_target),
                "gender": np.random.choice((1, 2, 3, 4), n_target),
                "id": range(n_target),
                "weight": pd.Series((1,) * n_target),
            }
        )
        target_df["is_respondent"] = False
        input_dataset = pd.concat([sample_df, target_df])

        with tempfile.TemporaryDirectory() as temp_dir, tempfile.NamedTemporaryFile(
            "w", suffix=".csv", delete=False
        ) as input_file:
            input_dataset.to_csv(path_or_buf=input_file)
            input_file.close()
            output_file = os.path.join(temp_dir, "weights_out.csv")
            diagnostics_output_file = os.path.join(temp_dir, "diagnostics_out.csv")
            features = "age,gender"

            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    input_file.name,
                    "--output_file",
                    output_file,
                    "--diagnostics_output_file",
                    diagnostics_output_file,
                    "--covariate_columns",
                    features,
                    "--method=cbps",
                ]
            )
            # run cli
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()
            # get the files created from cli to pandas to check them
            diagnostics_output = pd.read_csv(diagnostics_output_file, sep=",")
            self.assertEqual(
                diagnostics_output[diagnostics_output["metric"] == "adjustment_method"][
                    "var"
                ].values,
                np.array(["cbps"]),
            )

            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    input_file.name,
                    "--output_file",
                    output_file,
                    "--diagnostics_output_file",
                    diagnostics_output_file,
                    "--covariate_columns",
                    features,
                    "--method=ipw",
                    "--max_de=1.5",
                ]
            )
            # run cli
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()
            # get the files created from cli to pandas to check them
            diagnostics_output = pd.read_csv(diagnostics_output_file, sep=",")
            self.assertEqual(
                diagnostics_output[diagnostics_output["metric"] == "adjustment_method"][
                    "var"
                ].values,
                np.array(["ipw"]),
            )

    def test_method_works_with_rake(self):
        # TODO: ideally we'll have the example outside, and a different function for each of the methods (ipw, cbps, raking)
        np.random.seed(2021)
        n_sample = 1000
        n_target = 2000
        sample_df = pd.DataFrame(
            {
                "age": np.random.uniform(0, 100, n_sample),
                "gender": np.random.choice((1, 2, 3, 4), n_sample),
                "id": range(n_sample),
                "weight": pd.Series((1,) * n_sample),
            }
        )
        sample_df["is_respondent"] = True
        target_df = pd.DataFrame(
            {
                "age": np.random.uniform(0, 100, n_target),
                "gender": np.random.choice((1, 2, 3, 4), n_target),
                "id": range(n_target),
                "weight": pd.Series((1,) * n_target),
            }
        )
        target_df["is_respondent"] = False
        input_dataset = pd.concat([sample_df, target_df])

        with tempfile.TemporaryDirectory() as temp_dir, tempfile.NamedTemporaryFile(
            "w", suffix=".csv", delete=False
        ) as input_file:
            input_dataset.to_csv(path_or_buf=input_file)
            input_file.close()
            output_file = os.path.join(temp_dir, "weights_out.csv")
            diagnostics_output_file = os.path.join(temp_dir, "diagnostics_out.csv")
            features = "age,gender"

            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    input_file.name,
                    "--output_file",
                    output_file,
                    "--diagnostics_output_file",
                    diagnostics_output_file,
                    "--covariate_columns",
                    features,
                    "--method=rake",
                ]
            )
            # run cli
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()
            # get the files created from cli to pandas to check them
            diagnostics_output = pd.read_csv(diagnostics_output_file, sep=",")
            self.assertEqual(
                diagnostics_output[diagnostics_output["metric"] == "adjustment_method"][
                    "var"
                ].values,
                np.array(["rake"]),
            )

    def test_one_hot_encoding_works(self):
        with tempfile.TemporaryDirectory() as temp_dir, tempfile.NamedTemporaryFile(
            "w", suffix=".tsv", delete=False
        ) as in_file:

            # Assert value is False when "False" is passed
            out_file = os.path.join(temp_dir, "out.csv")
            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    in_file.name,
                    "--output_file",
                    out_file,
                    "--covariate_columns",
                    "x,y",
                    "--one_hot_encoding",
                    "False",
                ]
            )
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            self.assertFalse(cli.one_hot_encoding())
            self.assertEqual(type(cli.one_hot_encoding()), bool)

            # Assert value is True by default
            args2 = parser.parse_args(
                [
                    "--input_file",
                    in_file.name,
                    "--output_file",
                    out_file,
                    "--covariate_columns",
                    "x,y",
                ]
            )
            cli2 = BalanceCLI(args2)
            cli2.update_attributes_for_main_used_by_adjust()
            self.assertTrue(cli2.one_hot_encoding())
            self.assertEqual(type(cli2.one_hot_encoding()), bool)

            # Assert invalid value raises an error
            with self.assertRaises(ValueError):
                args3 = parser.parse_args(
                    [
                        "--input_file",
                        in_file.name,
                        "--output_file",
                        out_file,
                        "--covariate_columns",
                        "x,y",
                        "--one_hot_encoding",
                        "Invalid Value",
                    ]
                )
                cli3 = BalanceCLI(args3)
                cli3.update_attributes_for_main_used_by_adjust()

    def test_transformations_works(self):
        # TODO: ideally we'll have the example outside
        np.random.seed(2021)
        n_sample = 1000
        n_target = 2000
        sample_df = pd.DataFrame(
            {
                "age": np.random.uniform(0, 100, n_sample),
                "gender": np.random.choice((1, 2, 3, 4), n_sample),
                "id": range(n_sample),
                "weight": pd.Series((1,) * n_sample),
            }
        )
        sample_df["is_respondent"] = True
        target_df = pd.DataFrame(
            {
                "age": np.random.uniform(0, 100, n_target),
                "gender": np.random.choice((1, 2, 3, 4), n_target),
                "id": range(n_target),
                "weight": pd.Series((1,) * n_target),
            }
        )
        target_df["is_respondent"] = False
        input_dataset = pd.concat([sample_df, target_df])

        with tempfile.TemporaryDirectory() as temp_dir, tempfile.NamedTemporaryFile(
            "w", suffix=".csv", delete=False
        ) as input_file:
            input_dataset.to_csv(path_or_buf=input_file)
            input_file.close()
            output_file = os.path.join(temp_dir, "weights_out.csv")
            diagnostics_output_file = os.path.join(temp_dir, "diagnostics_out.csv")
            features = "age,gender"

            # test transformations=None
            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    input_file.name,
                    "--output_file",
                    output_file,
                    "--diagnostics_output_file",
                    diagnostics_output_file,
                    "--covariate_columns",
                    features,
                    "--transformations=None",
                ]
            )
            # run cli
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()
            # get the files created from cli to pandas to check them
            diagnostics_output = pd.read_csv(diagnostics_output_file, sep=",")
            self.assertEqual(
                diagnostics_output[diagnostics_output["metric"] == "model_coef"][
                    "var"
                ].values,
                np.array(["intercept", "age", "gender"]),
            )

            # test transformations='default'
            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    input_file.name,
                    "--output_file",
                    output_file,
                    "--diagnostics_output_file",
                    diagnostics_output_file,
                    "--covariate_columns",
                    features,
                    "--transformations=default",
                ]
            )
            # run cli
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()
            # get the files created from cli to pandas to check them
            diagnostics_output = pd.read_csv(diagnostics_output_file, sep=",")
            self.assertEqual(
                diagnostics_output[diagnostics_output["metric"] == "model_coef"][
                    "var"
                ].values,
                np.array(
                    [
                        "intercept",
                        "C(age, one_hot_encoding_greater_2)[(0.00264, 10.925]]",
                        "C(age, one_hot_encoding_greater_2)[(10.925, 20.624]]",
                        "C(age, one_hot_encoding_greater_2)[(20.624, 30.985]]",
                        "C(age, one_hot_encoding_greater_2)[(30.985, 41.204]]",
                        "C(age, one_hot_encoding_greater_2)[(41.204, 51.335]]",
                        "C(age, one_hot_encoding_greater_2)[(51.335, 61.535]]",
                        "C(age, one_hot_encoding_greater_2)[(61.535, 71.696]]",
                        "C(age, one_hot_encoding_greater_2)[(71.696, 80.08]]",
                        "C(age, one_hot_encoding_greater_2)[(80.08, 89.446]]",
                        "C(age, one_hot_encoding_greater_2)[(89.446, 99.992]]",
                        "C(gender, one_hot_encoding_greater_2)[(0.999, 2.0]]",
                        "C(gender, one_hot_encoding_greater_2)[(2.0, 3.0]]",
                        "C(gender, one_hot_encoding_greater_2)[(3.0, 4.0]]",
                    ]
                ),
            )

            # test default value for transformations
            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    input_file.name,
                    "--output_file",
                    output_file,
                    "--diagnostics_output_file",
                    diagnostics_output_file,
                    "--covariate_columns",
                    features,
                ]
            )
            # run cli
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()
            # get the files created from cli to pandas to check them
            diagnostics_output = pd.read_csv(diagnostics_output_file, sep=",")
            self.assertEqual(
                diagnostics_output[diagnostics_output["metric"] == "model_coef"][
                    "var"
                ].values,
                np.array(
                    [
                        "intercept",
                        "C(age, one_hot_encoding_greater_2)[(0.00264, 10.925]]",
                        "C(age, one_hot_encoding_greater_2)[(10.925, 20.624]]",
                        "C(age, one_hot_encoding_greater_2)[(20.624, 30.985]]",
                        "C(age, one_hot_encoding_greater_2)[(30.985, 41.204]]",
                        "C(age, one_hot_encoding_greater_2)[(41.204, 51.335]]",
                        "C(age, one_hot_encoding_greater_2)[(51.335, 61.535]]",
                        "C(age, one_hot_encoding_greater_2)[(61.535, 71.696]]",
                        "C(age, one_hot_encoding_greater_2)[(71.696, 80.08]]",
                        "C(age, one_hot_encoding_greater_2)[(80.08, 89.446]]",
                        "C(age, one_hot_encoding_greater_2)[(89.446, 99.992]]",
                        "C(gender, one_hot_encoding_greater_2)[(0.999, 2.0]]",
                        "C(gender, one_hot_encoding_greater_2)[(2.0, 3.0]]",
                        "C(gender, one_hot_encoding_greater_2)[(3.0, 4.0]]",
                    ]
                ),
            )

    def test_formula_works(self):
        # TODO: ideally we'll have the example outside
        np.random.seed(2021)
        n_sample = 1000
        n_target = 2000
        sample_df = pd.DataFrame(
            {
                "age": np.random.uniform(0, 100, n_sample),
                "gender": np.random.choice((1, 2, 3, 4), n_sample),
                "id": range(n_sample),
                "weight": pd.Series((1,) * n_sample),
            }
        )
        sample_df["is_respondent"] = True
        target_df = pd.DataFrame(
            {
                "age": np.random.uniform(0, 100, n_target),
                "gender": np.random.choice((1, 2, 3, 4), n_target),
                "id": range(n_target),
                "weight": pd.Series((1,) * n_target),
            }
        )
        target_df["is_respondent"] = False
        input_dataset = pd.concat([sample_df, target_df])

        with tempfile.TemporaryDirectory() as temp_dir, tempfile.NamedTemporaryFile(
            "w", suffix=".csv", delete=False
        ) as input_file:
            input_dataset.to_csv(path_or_buf=input_file)
            input_file.close()
            output_file = os.path.join(temp_dir, "weights_out.csv")
            diagnostics_output_file = os.path.join(temp_dir, "diagnostics_out.csv")
            features = "age,gender"

            # test no formula
            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    input_file.name,
                    "--output_file",
                    output_file,
                    "--diagnostics_output_file",
                    diagnostics_output_file,
                    "--covariate_columns",
                    features,
                    "--transformations=None",
                ]
            )
            # run cli
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()
            # get the files created from cli to pandas to check them
            diagnostics_output = pd.read_csv(diagnostics_output_file, sep=",")
            self.assertEqual(
                diagnostics_output[diagnostics_output["metric"] == "model_coef"][
                    "var"
                ].values,
                np.array(["intercept", "age", "gender"]),
            )

            # test transformations=age*gender
            parser = make_parser()
            args = parser.parse_args(
                [
                    "--input_file",
                    input_file.name,
                    "--output_file",
                    output_file,
                    "--diagnostics_output_file",
                    diagnostics_output_file,
                    "--covariate_columns",
                    features,
                    "--transformations=None",
                    "--formula=age*gender",
                ]
            )
            # run cli
            cli = BalanceCLI(args)
            cli.update_attributes_for_main_used_by_adjust()
            cli.main()
            # get the files created from cli to pandas to check them
            diagnostics_output = pd.read_csv(diagnostics_output_file, sep=",")
            self.assertEqual(
                diagnostics_output[diagnostics_output["metric"] == "model_coef"][
                    "var"
                ].values,
                np.array(["intercept", "age", "age:gender", "gender"]),
            )

    def test_cli_return_df_with_original_dtypes(self):
        out_True = check_some_flags(True, "--return_df_with_original_dtypes")
        out_False = check_some_flags(False, "--return_df_with_original_dtypes")

        # IF the flag 'return_df_with_original_dtypes' is not passed, then the dtypes of the output dataframes are different
        self.assertEqual(
            out_False["pd_in"].dtypes.to_dict(),
            {
                "x": dtype("float64"),
                "y": dtype("int64"),
                "is_respondent": dtype("int64"),
                "id": dtype("int64"),
                "weight": dtype("int64"),
            },
        )
        self.assertEqual(
            out_False["pd_out"].dtypes.to_dict(),
            {
                "id": dtype("int64"),
                "x": dtype("float64"),
                "y": dtype("float64"),
                "is_respondent": dtype("float64"),
                "weight": dtype("float64"),
            },
        )

        # IF the flag 'return_df_with_original_dtypes' IS passed, then the dtypes of the output dataframes are the SAME:
        # The input df dtypes are the same
        self.assertEqual(
            out_True["pd_in"].dtypes.to_dict(), out_False["pd_in"].dtypes.to_dict()
        )
        # But now the output df dtypes are also the same (the order of the columns is different though)
        self.assertEqual(
            out_True["pd_out"].dtypes.to_dict(),
            {
                "id": dtype("int64"),
                "x": dtype("float64"),
                "y": dtype("int64"),
                "is_respondent": dtype("int64"),
                "weight": dtype("int64"),
            },
        )

    def test_cli_standardize_types(self):

        out_False = check_some_flags(True, "--standardize_types=False")
        out_True = check_some_flags(
            False, "--standardize_types=True"
        )  # This is the same as not using the flag at all
        out_True_flag_on = check_some_flags(
            True, "--standardize_types=True"
        )  # This is the same as not using the flag at all

        # IF the flag 'skip_standardize_types' is not passed, then the dtypes of the output dataframes are different
        self.assertEqual(
            out_True["pd_in"].dtypes.to_dict(),
            {
                "x": dtype("float64"),
                "y": dtype("int64"),
                "is_respondent": dtype("int64"),
                "id": dtype("int64"),
                "weight": dtype("int64"),
            },
        )
        self.assertEqual(
            out_True["pd_out"].dtypes.to_dict(),
            {
                "id": dtype("int64"),
                "x": dtype("float64"),
                "y": dtype("float64"),
                "is_respondent": dtype("float64"),
                "weight": dtype("float64"),
            },
        )

        # IF the flag 'skip_standardize_types' IS passed, then the dtypes of the output dataframes are the SAME:
        # The input df dtypes are the same
        self.assertEqual(
            out_False["pd_in"].dtypes.to_dict(), out_True["pd_in"].dtypes.to_dict()
        )
        # But now the output df dtypes are also the same (the order of the columns is different though)
        self.assertEqual(
            out_False["pd_out"].dtypes.to_dict(),
            {
                "id": dtype("int64"),
                "x": dtype("float64"),
                "y": dtype("int64"),
                "is_respondent": dtype("int64"),
                "weight": dtype("float64"),
            },
        )

        self.assertEqual(
            out_True["pd_out"].dtypes.to_dict(),
            out_True_flag_on["pd_out"].dtypes.to_dict(),
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

import os
import sys

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "balance"
copyright = "Copyright © 2022 Meta Platforms, Inc."
author = "Meta Platforms"

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath(".."))

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",  # HTML generation from Py docstrings
    "sphinx.ext.autosummary",  # Generate Py docs recursively
    "sphinx.ext.napoleon",  # Support Google style docstrings
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
]
autosummary_generate = True  # Turn on sphinx.ext.autosummary
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "basic"
html_static_path = ["_static"]
